@@ -61,8 +61,24 @@ function logUpdated(...args) {
 }
 
 export function startListen() {
+  const targetWindow = Tabs.getWindow();
+
   browser.tabs.onActivated.addListener(onActivated);
-  browser.tabs.onUpdated.addListener(onUpdated);
+
+  let hasOnUpdated = false;
+  try {
+    if (typeof targetWindow === 'number') {
+      browser.tabs.onUpdated.addListener(onUpdated, { windowId: targetWindow });
+      hasOnUpdated = true;
+    }
+  }
+  catch (_error) {
+    /* browser.tabs.onUpdated filter not supported (Firefox 60 or earlier) */
+  }
+  if (!hasOnUpdated)
+    browser.tabs.onUpdated.addListener(onUpdated);
+
+  browser.tabs.onHighlighted.addListener(onHighlighted);
   browser.tabs.onCreated.addListener(onCreated);
   browser.tabs.onRemoved.addListener(onRemoved);
   browser.tabs.onMoved.addListener(onMoved);
@@ -74,6 +90,7 @@ export function startListen() {
 export function endListen() {
   browser.tabs.onActivated.removeListener(onActivated);
   browser.tabs.onUpdated.removeListener(onUpdated);
+  browser.tabs.onHighlighted.removeListener(onHighlighted);
   browser.tabs.onCreated.removeListener(onCreated);
   browser.tabs.onRemoved.removeListener(onRemoved);
   browser.tabs.onMoved.removeListener(onMoved);
@@ -112,7 +129,7 @@ function getOrBuildTabsContainer(hint) {
   return container;
 }
 
-const mLastClosedWhileActiveResolvers = new WeakMap();
+const mLastClosedWhileActiveResolvers = new WeakMap(); // used only on Firefox 64 and older
 
 async function onActivated(activeInfo) {
   const targetWindow = Tabs.getWindow();
@@ -126,15 +143,15 @@ async function onActivated(activeInfo) {
   try {
     const container = getOrBuildTabsContainer(activeInfo.windowId);
 
-    let byInternalOperation = parseInt(container.dataset.internalFocusCount) > 0;
+    let byInternalOperation = container.internalFocusCount > 0;
     if (byInternalOperation)
-      TabsContainer.decrementCounter(container, 'internalFocusCount');
-    const silently = parseInt(container.dataset.internalSilentlyFocusCount) > 0;
+      container.internalFocusCount--;
+    const silently = container.internalSilentlyFocusCount > 0;
     if (silently)
-      TabsContainer.decrementCounter(container, 'internalSilentlyFocusCount');
+      container.internalSilentlyFocusCount--;
     const byTabDuplication = parseInt(container.dataset.duplicatingTabsCount) > 0;
 
-    if (Tabs.hasCreatingTab())
+    if (Tabs.hasCreatingTab(activeInfo.windowId))
       await Tabs.waitUntilTabsAreCreated(activeInfo.tabId);
 
     const newTab = Tabs.getTabById({ tab: activeInfo.tabId, window: activeInfo.windowId });
@@ -144,27 +161,30 @@ async function onActivated(activeInfo) {
     }
 
     log('tabs.onActivated: ', dumpTab(newTab));
-    const oldActiveTabs = TabsInternalOperation.setTabFocused(newTab);
+    const oldActiveTabs = TabsInternalOperation.setTabActive(newTab);
 
-    let byCurrentTabRemove = mLastClosedWhileActiveResolvers.has(container);
-    if (byCurrentTabRemove) {
-      TabsContainer.incrementCounter(container, 'tryingReforcusForClosingCurrentTabCount');
-      mLastClosedWhileActiveResolvers.get(container)();
-      delete mLastClosedWhileActiveResolvers.delete(container);
-      const focusRedirected = await container.focusRedirectedForClosingCurrentTab;
-      delete container.focusRedirectedForClosingCurrentTab;
-      if (parseInt(container.dataset.tryingReforcusForClosingCurrentTabCount) > 0) // reduce count even if not redirected
-        TabsContainer.decrementCounter(container, 'tryingReforcusForClosingCurrentTabCount');
-      log('focusRedirected: ', focusRedirected);
-      if (focusRedirected) {
-        onCompleted();
-        return;
+    let byActiveTabRemove = !activeInfo.previousTabId;
+    if (!('successorTabId' in newTab.apiTab)) { // on Firefox 64 or older
+      byActiveTabRemove = mLastClosedWhileActiveResolvers.has(container);
+      if (byActiveTabRemove) {
+        container.tryingReforcusForClosingActiveTabCount++;
+        mLastClosedWhileActiveResolvers.get(container)();
+        delete mLastClosedWhileActiveResolvers.delete(container);
+        const focusRedirected = await container.focusRedirectedForClosingActiveTab;
+        delete container.focusRedirectedForClosingActiveTab;
+        if (container.tryingReforcusForClosingActiveTabCount > 0) // reduce count even if not redirected
+          container.tryingReforcusForClosingActiveTabCount--;
+        log('focusRedirected: ', focusRedirected);
+        if (focusRedirected) {
+          onCompleted();
+          return;
+        }
+      }
+      else if (container.tryingReforcusForClosingActiveTabCount > 0) { // treat as "redirected unintentional tab focus"
+        container.tryingReforcusForClosingActiveTabCount--;
+        byActiveTabRemove  = true;
+        byInternalOperation = false;
       }
-    }
-    else if (parseInt(container.dataset.tryingReforcusForClosingCurrentTabCount) > 0) { // treat as "redirected unintentional tab focus"
-      TabsContainer.decrementCounter(container, 'tryingReforcusForClosingCurrentTabCount');
-      byCurrentTabRemove  = true;
-      byInternalOperation = false;
     }
 
     if (!Tabs.ensureLivingTab(newTab)) { // it can be removed while waiting
@@ -172,12 +192,12 @@ async function onActivated(activeInfo) {
       return;
     }
 
-    let focusOverridden = Tabs.onActivating.dispatch(newTab, {
-      byCurrentTabRemove,
+    let focusOverridden = Tabs.onActivating.dispatch(newTab, Object.assign({}, activeInfo, {
+      byActiveTabRemove,
       byTabDuplication,
       byInternalOperation,
       silently
-    });
+    }));
     // don't do await if not needed, to process things synchronously
     if (focusOverridden instanceof Promise)
       focusOverridden = await focusOverridden;
@@ -192,13 +212,13 @@ async function onActivated(activeInfo) {
       return;
     }
 
-    const onActivatedReuslt = Tabs.onActivated.dispatch(newTab, {
+    const onActivatedReuslt = Tabs.onActivated.dispatch(newTab, Object.assign({}, activeInfo, {
       oldActiveTabs,
-      byCurrentTabRemove,
+      byActiveTabRemove,
       byTabDuplication,
       byInternalOperation,
       silently
-    });
+    }));
     // don't do await if not needed, to process things synchronously
     if (onActivatedReuslt instanceof Promise)
       await onActivatedReuslt;
@@ -218,7 +238,7 @@ async function onUpdated(tabId, changeInfo, tab) {
   TabIdFixer.fixTab(tab);
   tabId = tab.id;
 
-  if (Tabs.hasCreatingTab())
+  if (Tabs.hasCreatingTab(tab.windowId))
     await Tabs.waitUntilTabsAreCreated(tabId);
 
   const [onCompleted, previous] = addTabOperationQueue();
@@ -246,7 +266,21 @@ async function onUpdated(tabId, changeInfo, tab) {
       outside of TST (in other words, by any other addon.)
     */
     for (const key of Object.keys(changeInfo)) {
-      updatedTab.apiTab[key] = changeInfo[key];
+      if (key != 'index')
+        updatedTab.apiTab[key] = changeInfo[key];
+    }
+    if (changeInfo.url ||
+        changeInfo.status == 'complete') {
+      // On some edge cases internally changed "favIconUrl" is not
+      // notified, so we need to check actual favIconUrl manually.
+      // Known cases are:
+      //  * Transition from "about:privatebrowsing" to "about:blank"
+      //    https://github.com/piroor/treestyletab/issues/1916
+      //  * Reopen tab by Ctrl-Shift-T
+      browser.tabs.get(tabId).then(apiTab => {
+        if (apiTab.favIconUrl != updatedTab.apiTab.favIconUrl)
+          onUpdated(tabId, { favIconUrl: apiTab.favIconUrl }, apiTab);
+      });
     }
     if (configs.enableWorkaroundForBug1409262 &&
         tab.openerTabId != updatedTab.apiTab.TSTUpdatedOpenerTabId) {
@@ -254,9 +288,7 @@ async function onUpdated(tabId, changeInfo, tab) {
       updatedTab.apiTab.TSTUpdatedOpenerTabId = updatedTab.apiTab.openerTabId = tab.openerTabId;
     }
 
-    TabsUpdate.updateTab(updatedTab, changeInfo, {
-      tab: tab
-    });
+    TabsUpdate.updateTab(updatedTab, changeInfo, { tab });
     TabsUpdate.updateParentTab(Tabs.getParentTab(updatedTab));
 
     const onUpdatedResult = Tabs.onUpdated.dispatch(updatedTab, changeInfo);
@@ -271,6 +303,26 @@ async function onUpdated(tabId, changeInfo, tab) {
   }
 }
 
+const mTabsHighlightedTimers = new Map();
+const mLastHighlightedCount  = new Map();
+function onHighlighted(highlightInfo) {
+  let timer = mTabsHighlightedTimers.get(highlightInfo.windowId);
+  if (timer)
+    clearTimeout(timer);
+  if ((mLastHighlightedCount.get(highlightInfo.windowId) || 0) <= 1 &&
+      highlightInfo.tabIds.length == 1) {
+    // simple active tab switching
+    TabsUpdate.updateTabsHighlighted(highlightInfo);
+    return;
+  }
+  timer = setTimeout(() => {
+    mTabsHighlightedTimers.delete(highlightInfo.windowId);
+    TabsUpdate.updateTabsHighlighted(highlightInfo);
+    mLastHighlightedCount.set(highlightInfo.windowId, highlightInfo.tabIds.length);
+  }, 50);
+  mTabsHighlightedTimers.set(highlightInfo.windowId, timer);
+}
+
 function onCreated(tab) {
   const targetWindow = Tabs.getWindow();
   if (targetWindow && tab.windowId != targetWindow)
@@ -285,38 +337,54 @@ async function onNewTabTracked(tab) {
   if (targetWindow && tab.windowId != targetWindow)
     return null;
 
-  log('onNewTabTracked: ', tab);
+  log(`onNewTabTracked(id=${tab.id}): `, tab);
 
   const container = getOrBuildTabsContainer(tab.windowId);
-  const openedWithPosition   = parseInt(container.dataset.toBeOpenedTabsWithPositions) > 0;
-  const duplicatedInternally = parseInt(container.dataset.duplicatingTabsCount) > 0;
-  const maybeOrphan          = parseInt(container.dataset.toBeOpenedOrphanTabs) > 0;
-  const activeTab            = Tabs.getCurrentTab(container);
+  const positionedBySelf     = container.toBeOpenedTabsWithPositions > 0;
+  const duplicatedInternally = container.duplicatingTabsCount > 0;
+  const maybeOrphan          = container.toBeOpenedOrphanTabs > 0;
+  const activeTab            = Tabs.getActiveTab(container);
 
   Tabs.onBeforeCreate.dispatch(tab, {
-    maybeOpenedWithPosition: openedWithPosition,
+    positionedBySelf,
     maybeOrphan,
     activeTab
   });
 
-  if (Tabs.hasCreatingTab())
-    await Tabs.waitUntilAllTabsAreCreated();
+  if (Tabs.hasCreatingTab(tab.windowId))
+    await Tabs.waitUntilAllTabsAreCreated(tab.windowId);
 
   const [onCompleted, previous] = addTabOperationQueue();
   if (!configs.acceleratedTabOperations && previous)
     await previous;
 
-  log('onNewTabTracked: start to create tab for ', tab);
+  log(`onNewTabTracked(id=${tab.id}): start to create tab element`);
 
   try {
-    const hasNextTab = !!Tabs.getAllTabs(container)[tab.index];
-
     const newTab = Tabs.buildTab(tab, { inRemote: !!targetWindow });
     newTab.classList.add(Constants.kTAB_STATE_OPENING);
 
+    // New tab's index can become invalid because the value of "index" is same to
+    // the one given to browser.tabs.create() instead of actual index.
+    // See also: https://github.com/piroor/treestyletab/issues/2131
+    tab.index = Math.max(0, Math.min(tab.index, container.childNodes.length));
+
     const nextTab = Tabs.getAllTabs(container)[tab.index];
     container.insertBefore(newTab, nextTab);
 
+    if (nextTab)
+      reindexFollowingTabs(nextTab, tab.index + 1);
+
+    // We need to update "active" state of a new active tab immediately.
+    // Attaching of initial child tab (this new tab may become it) to an
+    // existing tab may produce collapsing of existing tree, and a
+    // collapsing tree may have the old active tab. On such cases TST
+    // tries to move focus to a nearest visible ancestor, instead of this
+    // new active tab.
+    // See also: https://github.com/piroor/treestyletab/issues/2155
+    if (tab.active)
+      TabsInternalOperation.setTabActive(newTab);
+
     const onTabCreatedInner = Tabs.addCreatingTab(newTab);
     const onTabCreated = (uniqueId) => { onTabCreatedInner(uniqueId); onCompleted(); };
     const uniqueId = await newTab.uniqueId;
@@ -334,12 +402,12 @@ async function onNewTabTracked(tab) {
     // tabs can be removed and detached while waiting, so cache them here for `detectTabActionFromNewPosition()`.
     const treeForActionDetection = Tabs.snapshotTreeForActionDetection(newTab);
 
-    if (openedWithPosition)
-      TabsContainer.decrementCounter(container, 'toBeOpenedTabsWithPositions');
+    if (positionedBySelf)
+      container.toBeOpenedTabsWithPositions--;
     if (maybeOrphan)
-      TabsContainer.decrementCounter(container, 'toBeOpenedOrphanTabs');
+      container.toBeOpenedOrphanTabs--;
     if (duplicatedInternally)
-      TabsContainer.decrementCounter(container, 'duplicatingTabsCount');
+      container.duplicatingTabsCount--;
 
     const duplicated = duplicatedInternally || uniqueId.duplicated;
     const restored   = uniqueId.restored;
@@ -347,7 +415,7 @@ async function onNewTabTracked(tab) {
       container.restoredCount = container.restoredCount || 0;
       container.restoredCount++;
       if (!container.allTabsRestored) {
-        log('Maybe starting to restore window ', tab.id);
+        log(`onNewTabTracked(id=${tab.id}): Maybe starting to restore window`);
         container.allTabsRestored = new Promise((resolve, _aReject) => {
           let lastCount = container.restoredCount;
           const timer = setInterval(() => {
@@ -366,17 +434,17 @@ async function onNewTabTracked(tab) {
       }
       Tabs.onRestoring.dispatch(newTab);
       await container.allTabsRestored;
-      log('onNewTabTracked: continued for restored tab ', tab.id);
+      log(`onNewTabTracked(id=${tab.id}): continued for restored tab`);
     }
     if (!container.parentNode ||
         !newTab.parentNode) {
-      log(' => aborted ', tab.id);
+      log(`onNewTabTracked(id=${tab.id}):  => aborted`);
       onTabCreated(uniqueId);
       return;
     }
 
     let moved = Tabs.onCreating.dispatch(newTab, {
-      maybeOpenedWithPosition: openedWithPosition,
+      positionedBySelf,
       maybeOrphan,
       restored,
       duplicated,
@@ -387,13 +455,14 @@ async function onNewTabTracked(tab) {
     if (moved instanceof Promise)
       moved = await moved;
     moved = moved === false;
+    log(`onNewTabTracked(id=${tab.id}): moved = `, moved);
 
     if (container.parentNode) { // it can be removed while waiting
-      TabsContainer.incrementCounter(container, 'openingCount');
+      container.openingTabs.add(tab.id);
       setTimeout(() => {
         if (!container.parentNode) // it can be removed while waiting
           return;
-        TabsContainer.decrementCounter(container, 'openingCount');
+        container.openingTabs.delete(tab.id);
       }, 0);
     }
 
@@ -402,11 +471,11 @@ async function onNewTabTracked(tab) {
       return;
     }
 
-    log('uniqueId: ', uniqueId);
+    log(`onNewTabTracked(id=${tab.id}): uniqueId = `, uniqueId);
 
     Tabs.onCreated.dispatch(newTab, {
-      openedWithPosition: openedWithPosition || moved,
-      skipFixupTree: !hasNextTab,
+      positionedBySelf: positionedBySelf || moved,
+      skipFixupTree: !nextTab,
       restored,
       duplicated,
       duplicatedInternally,
@@ -425,11 +494,27 @@ async function onNewTabTracked(tab) {
       checkRecycledTab(container);
     }
 
-    if (tab.active &&
-        Tabs.getCurrentTabs().some(tabElement => tabElement != newTab && tabElement.parentNode == newTab.parentNode))
-      onActivated({ tabId: tab.id, windowId: tab.windowId });
-
     onTabCreated(uniqueId);
+
+    // tab can be changed while creating!
+    const renewedTab = await browser.tabs.get(tab.id);
+    const changedProps = {};
+    for (const key of Object.keys(renewedTab)) {
+      if (tab[key] != renewedTab[key])
+        changedProps[key] = renewedTab[key];
+    }
+    if (Object.keys(renewedTab).length > 0)
+      onUpdated(tab.id, changedProps, renewedTab);
+
+    const currentActiveTab = Tabs.getActiveTabs().find(tabElement => tabElement != newTab && tabElement.parentNode == newTab.parentNode);
+    if (renewedTab.active &&
+        currentActiveTab)
+      onActivated({
+        tabId:         tab.id,
+        windowId:      tab.windowId,
+        previousTabId: currentActiveTab.apiTab.id
+      });
+
     return newTab;
   }
   catch(e) {
@@ -438,6 +523,16 @@ async function onNewTabTracked(tab) {
   }
 }
 
+function reindexFollowingTabs(startTab, startIndex) {
+  log('reindexFollowingTabs ', { startTab, startIndex });
+  let followingTab = startTab;
+  let newIndex = startIndex;
+  while (followingTab) {
+    followingTab.apiTab.index = newIndex++;
+    followingTab = Tabs.getNextTab(followingTab);
+  }
+}
+
 // "Recycled tab" is an existing but reused tab for session restoration.
 const kBASE_RECYCLED_TAB_CONDITION = `li:not(.${Constants.kTAB_STATE_RESTORED}):not(.${Constants.kTAB_STATE_OPENING})`;
 function checkRecycledTab(container) {
@@ -473,12 +568,12 @@ async function onRemoved(tabId, removeInfo) {
     return;
 
   const container = getOrBuildTabsContainer(removeInfo.windowId);
-  const byInternalOperation = parseInt(container.dataset.internalClosingCount) > 0;
+  const byInternalOperation = container.internalClosingTabs.has(tabId);
   if (byInternalOperation)
-    TabsContainer.decrementCounter(container, 'internalClosingCount');
+    container.internalClosingTabs.delete(tabId);
 
-  if (Tabs.hasCreatingTab())
-    await Tabs.waitUntilAllTabsAreCreated();
+  if (Tabs.hasCreatingTab(removeInfo.windowId))
+    await Tabs.waitUntilAllTabsAreCreated(removeInfo.windowId);
 
   const [onCompleted, previous] = addTabOperationQueue();
   if (!configs.acceleratedTabOperations && previous)
@@ -495,25 +590,34 @@ async function onRemoved(tabId, removeInfo) {
 
     Tabs.onStateChanged.dispatch(oldTab);
 
-    if (Tabs.isActive(oldTab)) {
+    if (Tabs.isActive(oldTab) &&
+        !('successorTabId' in oldTab.apiTab)) { // on Firefox 64 or older
       const resolver = Tabs.fetchClosedWhileActiveResolver(oldTab);
       if (resolver)
         mLastClosedWhileActiveResolvers.set(container, resolver);
     }
 
-    const onRemovingResult = Tabs.onRemoving.dispatch(oldTab, {
+    const onRemovingResult = Tabs.onRemoving.dispatch(oldTab, Object.assign({}, removeInfo, {
       byInternalOperation
-    });
+    }));
     // don't do await if not needed, to process things synchronously
     if (onRemovingResult instanceof Promise)
       await onRemovingResult;
 
+    // The removing tab may be attached to tree/someone attached to the removing tab.
+    // We need to clear them by onRemoved handlers.
+    const oldChildren = Tabs.getChildTabs(oldTab);
+    const oldParent   = Tabs.getParentTab(oldTab);
     oldTab[Constants.kTAB_STATE_REMOVING] = true;
     oldTab.classList.add(Constants.kTAB_STATE_REMOVING);
 
-    const onRemovedReuslt = Tabs.onRemoved.dispatch(oldTab, {
-      byInternalOperation
-    });
+    reindexFollowingTabs(Tabs.getNextTab(oldTab), oldTab.apiTab.index);
+
+    const onRemovedReuslt = Tabs.onRemoved.dispatch(oldTab, Object.assign({}, removeInfo, {
+      byInternalOperation,
+      oldChildren,
+      oldParent
+    }));
     // don't do await if not needed, to process things synchronously
     if (onRemovedReuslt instanceof Promise)
       await onRemovedReuslt;
@@ -554,19 +658,26 @@ async function onMoved(tabId, moveInfo) {
     return;
 
   const container = getOrBuildTabsContainer(moveInfo.windowId);
-  const byInternalOperation = parseInt(container.dataset.internalMovingCount) > 0;
 
-  if (Tabs.hasCreatingTab())
+  // Firefox may move the tab between TabsMove.moveTabsInternallyBefore/After()
+  // and TabsMove.syncTabsPositionToApiTabs(). We should treat such a movement
+  // as an "internal" operation also, because we need to suppress "move back"
+  // and other fixup operations around tabs moved by foreign triggers, on such
+  // cases. Don't mind, the tab will be rearranged again by delayed
+  // TabsMove.syncTabsPositionToApiTabs() anyway!
+  const maybeInternalOperation = container.internalMovingTabs.has(tabId);
+
+  if (Tabs.hasCreatingTab(moveInfo.windowId))
     await Tabs.waitUntilTabsAreCreated(tabId);
-  if (Tabs.hasMovingTab())
-    await Tabs.waitUntilAllTabsAreMoved();
+  if (Tabs.hasMovingTab(moveInfo.windowId))
+    await Tabs.waitUntilAllTabsAreMoved(moveInfo.windowId);
 
   const [onCompleted, previous] = addTabOperationQueue();
   if (!configs.acceleratedTabOperations && previous)
     await previous;
 
   try {
-    const onTabMoved = Tabs.addMovingTabId(tabId);
+    const onTabMoved = Tabs.addMovingTabId(tabId, moveInfo.windowId);
     const completelyMoved = () => { onTabMoved(); onCompleted() };
 
     /* When a tab is pinned, tabs.onMoved may be notified before
@@ -577,8 +688,8 @@ async function onMoved(tabId, moveInfo) {
        do following processes after the tab is completely pinned. */
     const movedTab = Tabs.getTabById({ tab: tabId, window: moveInfo.windowId });
     if (!movedTab) {
-      if (byInternalOperation)
-        TabsContainer.decrementCounter(container, 'internalMovingCount');
+      if (maybeInternalOperation)
+        container.internalMovingTabs.delete(tabId);
       completelyMoved();
       return;
     }
@@ -590,19 +701,21 @@ async function onMoved(tabId, moveInfo) {
       oldPreviousTab = tabs[moveInfo.toIndex < moveInfo.fromIndex ? moveInfo.fromIndex : moveInfo.fromIndex - 1];
       oldNextTab     = tabs[moveInfo.toIndex < moveInfo.fromIndex ? moveInfo.fromIndex + 1 : moveInfo.fromIndex];
     }
+
+    let alreadyMoved = false;
+    if (container.alreadyMovedTabs.has(tabId)) {
+      container.alreadyMovedTabs.delete(tabId);
+      alreadyMoved = true;
+    }
+
     const extendedMoveInfo = Object.assign({}, moveInfo, {
-      byInternalOperation,
+      byInternalOperation: maybeInternalOperation,
+      alreadyMoved,
       oldPreviousTab,
       oldNextTab
     });
     log('tabs.onMoved: ', dumpTab(movedTab), extendedMoveInfo, movedTab.apiTab);
 
-    let alreadyMoved = false;
-    if (parseInt(container.dataset.alreadyMovedTabsCount) > 0) {
-      TabsContainer.decrementCounter(container, 'alreadyMovedTabsCount');
-      alreadyMoved = true;
-    }
-
     let canceled = Tabs.onMoving.dispatch(movedTab, extendedMoveInfo);
     // don't do await if not needed, to process things synchronously
     if (canceled instanceof Promise)
@@ -613,28 +726,29 @@ async function onMoved(tabId, moveInfo) {
       let newNextIndex = extendedMoveInfo.toIndex;
       if (extendedMoveInfo.fromIndex < newNextIndex)
         newNextIndex++;
-      const tabs    = Tabs.getAllTabs(movedTab);
-      const nextTab = tabs[newNextIndex];
-      if (!alreadyMoved && movedTab.nextSibling != nextTab) {
+      const nextTab = Tabs.getAllTabs(movedTab)[newNextIndex];
+      if (!alreadyMoved &&
+          Tabs.getNextTab(movedTab) != nextTab) {
         container.insertBefore(movedTab, nextTab);
+        const tabs       = Tabs.getAllTabs(movedTab);
+        const startIndex = Math.max(Math.min(moveInfo.fromIndex, moveInfo.toIndex), 0);
+        const endIndex   = Math.min(Math.max(moveInfo.fromIndex, moveInfo.toIndex), tabs.length - 1);
+        for (let i = startIndex; i <= endIndex; i++) {
+          tabs[i].apiTab.index = i;
+        }
         log('Tab nodes rearranged by tabs.onMoved listener:\n'+(!configs.debug ? '' :
-          Array.slice(container.childNodes)
-            .map(tab => tab.id+(tab == movedTab ? '[MOVED]' : ''))
-            .join('\n')
-            .replace(/^/gm, ' - ')));
-      }
-      const startIndex = Math.max(Math.min(moveInfo.fromIndex, moveInfo.toIndex), 0);
-      const endIndex   = Math.min(Math.max(moveInfo.fromIndex, moveInfo.toIndex), tabs.length - 1);
-      for (let i = startIndex; i < endIndex; i++) {
-        tabs[i].apiTab.index = i;
+          Array.from(container.childNodes)
+            .map(tab => ' - '+tab.apiTab.index+': '+tab.id+(tab == movedTab ? '[MOVED]' : ''))
+            .join('\n')),
+            { moveInfo, startIndex, endIndex });
       }
       const onMovedResult = Tabs.onMoved.dispatch(movedTab, extendedMoveInfo);
       // don't do await if not needed, to process things synchronously
       if (onMovedResult instanceof Promise)
         await onMovedResult;
     }
-    if (byInternalOperation)
-      TabsContainer.decrementCounter(container, 'internalMovingCount');
+    if (maybeInternalOperation)
+      container.internalMovingTabs.delete(tabId);
     completelyMoved();
   }
   catch(e) {
@@ -672,13 +786,13 @@ async function onAttached(tabId, attachInfo) {
     TabIdFixer.fixTab(apiTab);
 
     TabsInternalOperation.clearOldActiveStateInWindow(attachInfo.newWindowId);
-    const info = mTreeInfoForTabsMovingAcrossWindows.get(tabId);
+    const info = Object.assign({}, attachInfo, mTreeInfoForTabsMovingAcrossWindows.get(tabId));
     mTreeInfoForTabsMovingAcrossWindows.delete(tabId);
 
     const newTab = await onNewTabTracked(apiTab);
-    const byInternalOperation = newTab && parseInt(newTab.parentNode.dataset.toBeAttachedTabs) > 0;
+    const byInternalOperation = newTab && newTab.parentNode.toBeAttachedTabs.has(apiTab.id);
     if (byInternalOperation)
-      TabsContainer.decrementCounter(newTab.parentNode, 'toBeAttachedTabs');
+      newTab.parentNode.toBeAttachedTabs.delete(apiTab.id);
     info.byInternalOperation = info.byInternalOperation || byInternalOperation;
 
     if (!byInternalOperation) { // we should process only tabs attached by others.
@@ -713,15 +827,15 @@ async function onDetached(tabId, detachInfo) {
       return;
     }
 
-    const byInternalOperation = parseInt(oldTab.parentNode.dataset.toBeDetachedTabs) > 0;
+    const byInternalOperation = oldTab.parentNode.toBeDetachedTabs.has(tabId);
     if (byInternalOperation)
-      TabsContainer.decrementCounter(oldTab.parentNode, 'toBeDetachedTabs');
+      oldTab.parentNode.toBeDetachedTabs.delete(tabId);
 
-    const info = {
+    const info = Object.assign({}, detachInfo, {
       byInternalOperation,
       windowId:    detachInfo.oldWindowId,
       descendants: Tabs.getDescendantTabs(oldTab)
-    };
+    });
     mTreeInfoForTabsMovingAcrossWindows.set(tabId, info);
 
     Tabs.onStateChanged.dispatch(oldTab);
@@ -730,10 +844,17 @@ async function onDetached(tabId, detachInfo) {
       Tabs.onDetached.dispatch(oldTab, info);
 
     const container = oldTab.parentNode;
+    const nextTab   = oldTab.nextSibling;
+    const oldIndex  = oldTab.apiTab.index;
     clearTabRelationsForRemovedTab(oldTab);
     container.removeChild(oldTab);
-    if (!container.hasChildNodes())
+    if (container.hasChildNodes()) {
+      if (nextTab)
+        reindexFollowingTabs(nextTab, oldIndex);
+    }
+    else {
       container.parentNode.removeChild(container);
+    }
 
     onCompleted();
   }
@@ -744,6 +865,9 @@ async function onDetached(tabId, detachInfo) {
 }
 
 async function onWindowRemoved(windowId) {
+  mTabsHighlightedTimers.delete(windowId);
+  mLastHighlightedCount.delete(windowId);
+
   const [onCompleted, previous] = addTabOperationQueue();
   if (!configs.acceleratedTabOperations && previous)
     await previous;
