@@ -48,6 +48,7 @@ import * as Constants from '/common/constants.js';
 import * as Tabs from '/common/tabs.js';
 import * as TabsOpen from '/common/tabs-open.js';
 import * as Tree from '/common/tree.js';
+import * as Commands from '/common/commands.js';
 import * as TSTAPI from '/common/tst-api.js';
 import * as Scroll from './scroll.js';
 import * as EventUtils from './event-utils.js';
@@ -66,6 +67,7 @@ const kBOOKMARK_FOLDER  = 'x-moz-place:';
 const kDROP_BEFORE  = 'before';
 const kDROP_ON_SELF = 'self';
 const kDROP_AFTER   = 'after';
+const kDROP_IMPOSSIBLE = 'impossible';
 
 const kDROP_POSITION = 'data-drop-position';
 
@@ -76,10 +78,8 @@ let mLongHoverExpandedTabs = [];
 let mLongHoverTimer;
 let mLongHoverTimerNext;
 
-let mDelayedDragEnter;
-let mDelayedDragLeave;
-
 let mDraggingOnSelfWindow = false;
+let mDraggingOnDraggedTabs = false;
 
 let mCapturingForDragging = false;
 let mReadyToCaptureMouseEvents = false;
@@ -88,8 +88,10 @@ let mLastDropPosition      = null;
 let mDragTargetIsClosebox  = false;
 let mCurrentDragData       = null;
 
+let mDragBehaviorNotification;
+
 export function init() {
-  document.addEventListener('dragstart', onDragStart);
+  document.addEventListener('dragstart', onDragStart); // eslint-disable-line no-use-before-define
   document.addEventListener('dragover', onDragOver);
   document.addEventListener('dragenter', onDragEnter);
   document.addEventListener('dragleave', onDragLeave);
@@ -97,6 +99,8 @@ export function init() {
   document.addEventListener('drop', onDrop);
 
   browser.runtime.onMessage.addListener(onMessage);
+
+  mDragBehaviorNotification = document.getElementById('tab-drag-notification');
 }
 
 
@@ -104,14 +108,20 @@ export function isCapturingForDragging() {
   return mCapturingForDragging;
 }
 
-export function startMultiDrag(tab, aIsClosebox) {
-  TSTAPI.sendMessage({
+// for backward compatibility with Multiple Tab Handler 2.x on Firefox ESR60
+export async function legacyStartMultiDrag(tab, aIsClosebox) {
+  const windowId = Tabs.getWindow();
+  const results = await TSTAPI.sendMessage({
     type:   TSTAPI.kNOTIFY_TAB_DRAGREADY,
     tab:    TSTAPI.serializeTab(tab),
-    window: Tabs.getWindow(),
+    window: windowId,
+    windowId,
     startOnClosebox: aIsClosebox
   });
-  mReadyToCaptureMouseEvents = true;
+  if (results.every(result => result.result !== false)) {
+    mReadyToCaptureMouseEvents = true;
+  }
+  return results;
 }
 
 export function endMultiDrag(tab, aCoordinates) {
@@ -125,6 +135,7 @@ export function endMultiDrag(tab, aCoordinates) {
       type:    TSTAPI.kNOTIFY_TAB_DRAGEND,
       tab:     serializedTab,
       window:  tab && tab.apiTab.windowId,
+      windowId: tab && tab.apiTab.windowId,
       clientX: aCoordinates.clientX,
       clientY: aCoordinates.clientY
     });
@@ -136,6 +147,7 @@ export function endMultiDrag(tab, aCoordinates) {
       type:    TSTAPI.kNOTIFY_TAB_DRAGCANCEL,
       tab:     serializedTab,
       window:  tab && tab.apiTab.windowId,
+      windowId: tab && tab.apiTab.windowId,
       clientX: aCoordinates.clientX,
       clientY: aCoordinates.clientY
     });
@@ -244,6 +256,9 @@ function getDropAction(event) {
     return info.targetTabs[info.targetTabs.length - 1];
   });
   info.defineGetter('canDrop', () => {
+    if (info.dropPosition == kDROP_IMPOSSIBLE)
+      return false;
+
     const draggedApiTab               = info.dragData && info.dragData.apiTab;
     const isPrivateBrowsingTabDragged = draggedApiTab && draggedApiTab.incognito;
     if (draggedApiTab &&
@@ -296,12 +311,20 @@ function getDropAction(event) {
       info.targetTab    = info.insertBefore = info.firstTargetTab;
       info.dropPosition = kDROP_BEFORE;
       info.action       = action;
+      if (info.draggedAPITab &&
+          !info.draggedAPITab.pinned &&
+          Tabs.isPinned(info.targetTab))
+        info.dropPosition = kDROP_IMPOSSIBLE;
     }
     else if (event.clientY > info.lastTargetTab.getBoundingClientRect().bottom) {
       //log('dragging below the last tab');
       info.targetTab    = info.insertAfter = info.lastTargetTab;
       info.dropPosition = kDROP_AFTER;
       info.action       = action;
+      if (info.draggedAPITab &&
+          info.draggedAPITab.pinned &&
+          !Tabs.isPinned(info.targetTab))
+        info.dropPosition = kDROP_IMPOSSIBLE;
     }
     return info;
   }
@@ -355,83 +378,43 @@ function getDropAction(event) {
         // if (info.insertBefore)
         //  log('insertBefore = ', dumpTab(info.insertBefore));
       });
+      if (info.draggedAPITab &&
+          info.draggedAPITab.pinned != Tabs.isPinned(targetTab))
+        info.dropPosition = kDROP_IMPOSSIBLE;
     }; break;
 
     case kDROP_BEFORE: {
       //log('drop position = before the tab');
-      /* strategy
-           +-----------------------------------------------------
-           |     <= detach from parent, and move
-           |[TARGET  ]
-           +-----------------------------------------------------
-           |  [      ]
-           |     <= attach to the parent of the target, and move
-           |[TARGET  ]
-           +-----------------------------------------------------
-           |[        ]
-           |     <= attach to the parent of the target, and move
-           |[TARGET  ]
-           +-----------------------------------------------------
-           |[        ]
-           |     <= attach to the parent of the target (previous tab), and move
-           |  [TARGET]
-           +-----------------------------------------------------
-      */
-      const prevTab = Tabs.getPreviousVisibleTab(targetTab);
-      if (!prevTab) {
-        // allow to drop pinned tab to beside of another pinned tab
-        if (info.draggedAPITab && info.draggedAPITab.pinned) {
-          info.action       = Constants.kACTION_MOVE;
-          info.insertBefore = targetTab;
-        }
-        else {
-          info.action       = Constants.kACTION_MOVE | Constants.kACTION_DETACH;
-          info.insertBefore = info.firstTargetTab;
-        }
-      }
-      else {
-        const prevLevel   = Number(prevTab.getAttribute(Constants.kLEVEL) || 0);
-        const targetLevel = Number(targetTab.getAttribute(Constants.kLEVEL) || 0);
-        info.parent       = (prevLevel < targetLevel) ? prevTab : Tabs.getParentTab(targetTab) ;
-        info.action       = Constants.kACTION_MOVE | (info.parent ? Constants.kACTION_ATTACH : Constants.kACTION_DETACH );
-        info.insertBefore = targetTab;
-      }
+      const referenceTabs = Tree.calculateReferenceTabsFromInsertionPosition(info.draggedTab, {
+        insertBefore: targetTab
+      });
+      if (referenceTabs.parent)
+        info.parent = referenceTabs.parent;
+      if (referenceTabs.insertBefore)
+        info.insertBefore = referenceTabs.insertBefore;
+      if (referenceTabs.insertAfter)
+        info.insertAfter = referenceTabs.insertAfter;
+      info.action = Constants.kACTION_MOVE | (info.parent ? Constants.kACTION_ATTACH : Constants.kACTION_DETACH );
       //if (info.insertBefore)
       //  log('insertBefore = ', dumpTab(info.insertBefore));
+      if (info.draggedAPITab &&
+          info.draggedAPITab.pinned != Tabs.isPinned(targetTab))
+        info.dropPosition = kDROP_IMPOSSIBLE;
     }; break;
 
     case kDROP_AFTER: {
       //log('drop position = after the tab');
-      /* strategy
-           +-----------------------------------------------------
-           |[TARGET  ]
-           |     <= if the target has a parent, attach to it and and move
-           +-----------------------------------------------------
-           |  [TARGET]
-           |     <= attach to the parent of the target, and move
-           |[        ]
-           +-----------------------------------------------------
-           |[TARGET  ]
-           |     <= attach to the parent of the target, and move
-           |[        ]
-           +-----------------------------------------------------
-           |[TARGET  ]
-           |     <= attach to the target, and move
-           |  [      ]
-           +-----------------------------------------------------
-      */
-      const nextTab = Tabs.getNextVisibleTab(targetTab);
-      if (!nextTab) {
-        info.action = Constants.kACTION_MOVE | Constants.kACTION_ATTACH;
-        info.parent = Tabs.getParentTab(targetTab);
-      }
-      else {
-        const targetLevel = Number(targetTab.getAttribute(Constants.kLEVEL) || 0);
-        const nextLevel   = Number(nextTab.getAttribute(Constants.kLEVEL) || 0);
-        info.parent       = (targetLevel < nextLevel) ? targetTab : Tabs.getParentTab(targetTab) ;
-        info.action       = Constants.kACTION_MOVE | (info.parent ? Constants.kACTION_ATTACH : Constants.kACTION_DETACH );
-        info.insertBefore = nextTab;
-        info.insertAfter  = targetTab;
+      const referenceTabs = Tree.calculateReferenceTabsFromInsertionPosition(info.draggedTab, {
+        insertAfter: targetTab
+      });
+      if (referenceTabs.parent)
+        info.parent = referenceTabs.parent;
+      if (referenceTabs.insertBefore)
+        info.insertBefore = referenceTabs.insertBefore;
+      if (referenceTabs.insertAfter)
+        info.insertAfter = referenceTabs.insertAfter;
+      info.action = Constants.kACTION_MOVE | (info.parent ? Constants.kACTION_ATTACH : Constants.kACTION_DETACH );
+      if (info.insertBefore) {
         /* strategy
              +-----------------------------------------------------
              |[TARGET   ]
@@ -440,7 +423,7 @@ function getDropAction(event) {
              +-----------------------------------------------------
         */
         if (info.draggedAPITab &&
-            info.draggedAPITab.id == nextTab.apiTab.id) {
+            info.draggedAPITab.id == info.insertBefore.apiTab.id) {
           info.action       = Constants.kACTION_MOVE | Constants.kACTION_ATTACH;
           info.parent       = Tabs.getParentTab(targetTab);
           info.defineGetter('insertBefore', () => {
@@ -459,6 +442,9 @@ function getDropAction(event) {
           });
         }
       }
+      if (info.draggedAPITab &&
+          info.draggedAPITab.pinned != Tabs.isPinned(Tabs.getNextVisibleTab(targetTab)))
+        info.dropPosition = kDROP_IMPOSSIBLE;
     }; break;
   }
 
@@ -497,7 +483,7 @@ function isDraggingAllTabs(tab, tabs) {
   return draggingTabs.length == (tabs || Tabs.getAllTabs(tab)).length;
 }
  
-function isDraggingAllCurrentTabs(tab) {
+function isDraggingAllActiveTabs(tab) {
   return isDraggingAllTabs(tab, Tabs.getAllTabs(tab));
 }
 
@@ -613,8 +599,8 @@ function retrieveURIsFromData(aData, type) {
         .replace(/\n\n+/g, '\n')
         .trim()
         .split('\n')
-        .filter(line => {
-          return /^\w+:\/\/.+/.test(line);
+        .map(line => {
+          return /^\w+:\/\/.+/.test(line) ? line : `about:treestyletab-search?${line}`;
         });
   }
   return [];
@@ -664,12 +650,18 @@ async function getDroppedLinksOnTabBehavior() {
 
 let mFinishCanceledDragOperation;
 
-function onDragStart(event) {
-  log('onDragStart: start ', event);
+export const onDragStart = EventUtils.wrapWithErrorHandler(function onDragStart(event, options = {}) {
+  log('onDragStart: start ', event, options);
   clearDraggingTabsState(); // clear previous state anyway
 
-  const dragData = getDragDataFromOneTab(event.target, {
-    shouldIgnoreDescendants: event.shiftKey
+  const behavior = 'behavior' in options ? options.behavior :
+    event.shiftKey ? configs.tabDragBehaviorShift :
+      configs.tabDragBehavior;
+  const shouldIgnoreDescendants = !(behavior & Constants.kDRAG_BEHAVIOR_WHOLE_TREE);
+  const allowBookmark           = !!(behavior & Constants.kDRAG_BEHAVIOR_ALLOW_BOOKMARK);
+
+  const dragData = getDragDataFromOneTab(options.target || event.target, {
+    shouldIgnoreDescendants
   });
   if (!dragData.tabNode) {
     log('onDragStart: canceled / no dragged tab from drag data');
@@ -687,10 +679,12 @@ function onDragStart(event) {
     const startOnClosebox = mDragTargetIsClosebox = mousedown.detail.closebox;
     if (startOnClosebox)
       mLastDragEnteredTarget = SidebarTabs.getClosebox(tab);
+    const windowId = Tabs.getWindow();
     TSTAPI.sendMessage({
       type:   TSTAPI.kNOTIFY_TAB_DRAGSTART,
       tab:    TSTAPI.serializeTab(tab),
-      window: Tabs.getWindow(),
+      window: windowId,
+      windowId,
       startOnClosebox
     });
     window.addEventListener('mouseover', onTSTAPIDragEnter, { capture: true });
@@ -703,8 +697,8 @@ function onDragStart(event) {
   EventUtils.cancelHandleMousedown();
 
   // dragging on clickable element will be expected to cancel the operation
-  if (EventUtils.isEventFiredOnClosebox(event) ||
-      EventUtils.isEventFiredOnClickable(event)) {
+  if (EventUtils.isEventFiredOnClosebox(options.target || event) ||
+      EventUtils.isEventFiredOnClickable(options.target || event)) {
     log('onDragStart: canceled / on undraggable element');
     event.stopPropagation();
     event.preventDefault();
@@ -712,6 +706,7 @@ function onDragStart(event) {
   }
 
   mDraggingOnSelfWindow = true;
+  mDraggingOnDraggedTabs = true;
   mLastDropPosition = null;
 
   const dt = event.dataTransfer;
@@ -730,16 +725,22 @@ function onDragStart(event) {
   });
 
   const mozUrl  = [];
+  const urlList = [];
   for (const draggedTab of dragData.tabNodes) {
     draggedTab.classList.add(Constants.kTAB_STATE_DRAGGING);
     mozUrl.push(`${draggedTab.apiTab.url}\n${draggedTab.apiTab.title}`);
+    urlList.push(`#${draggedTab.apiTab.title}\n${draggedTab.apiTab.url}`);
   }
-  let allowBookmark = configs.allowBookmarkCreationFromDraggedTree;
-  if (event.shiftKey)
-    allowBookmark = !allowBookmark;
   if (allowBookmark) {
-    // This is required to create bookmarks by drag and drop.
+    log('set kTYPE_X_MOZ_URL');
     dt.setData(kTYPE_X_MOZ_URL, mozUrl.join('\n'));
+    log('set kTYPE_URI_LIST');
+    dt.setData(kTYPE_URI_LIST, urlList.join('\n'));
+  }
+
+  if (options.target) {
+    const tabRect = options.target.getBoundingClientRect();
+    dt.setDragImage(options.target, event.clientX - tabRect.left, event.clientY - tabRect.top);
   }
 
   Tabs.getTabsContainer(tab).classList.add(kTABBAR_STATE_TAB_DRAGGING);
@@ -753,9 +754,48 @@ function onDragStart(event) {
   // See also: https://github.com/piroor/treestyletab/issues/1778#issuecomment-404569842
   mFinishCanceledDragOperation = setTimeout(finishDrag, 250);
 
+  if (!('behavior' in options) &&
+      configs.showTabDragBehaviorNotification) {
+    const currentBehavior = event.shiftKey ? configs.tabDragBehaviorShift : configs.tabDragBehavior;
+    const invertedBehavior = event.shiftKey ? configs.tabDragBehavior : configs.tabDragBehaviorShift;
+    const count            = dragData.tabNodes.length;
+    const currentResult    = getTabDragBehaviorNotificationMessageType(currentBehavior, count);
+    const invertedResult   = getTabDragBehaviorNotificationMessageType(invertedBehavior, count);
+    const invertSuffix     = event.shiftKey ? 'without_shift' : 'with_shift';
+    mDragBehaviorNotification.firstChild.textContent = [
+      browser.i18n.getMessage(`tabDragBehaviorNotification_message_base`, [
+        browser.i18n.getMessage(`tabDragBehaviorNotification_message_${currentResult}`)]),
+      browser.i18n.getMessage(`tabDragBehaviorNotification_message_inverted_base_${invertSuffix}`, [
+        browser.i18n.getMessage(`tabDragBehaviorNotification_message_${invertedResult}`)]),
+    ].join('\n');
+    mDragBehaviorNotification.firstChild.style.animationDuration = browser.i18n.getMessage('tabDragBehaviorNotification_message_duration');
+    mDragBehaviorNotification.classList.remove('hiding');
+    mDragBehaviorNotification.classList.add('shown');
+  }
+
+  TSTAPI.sendMessage({
+    type:     TSTAPI.kNOTIFY_NATIVE_TAB_DRAGSTART,
+    tab:      TSTAPI.serializeTab(tab),
+    windowId: Tabs.getWindow()
+  });
+
   log('onDragStart: started');
+});
+
+function getTabDragBehaviorNotificationMessageType(behavior, count) {
+  if (behavior & Constants.kDRAG_BEHAVIOR_WHOLE_TREE && count > 1) {
+    if (behavior & Constants.kDRAG_BEHAVIOR_ALLOW_BOOKMARK)
+      return 'tree_bookmark';
+    else
+      return 'tree_tearoff';
+  }
+  else {
+    if (behavior & Constants.kDRAG_BEHAVIOR_ALLOW_BOOKMARK)
+      return 'tab_bookmark';
+    else
+      return 'tab_tearoff';
+  }
 }
-onDragStart = EventUtils.wrapWithErrorHandler(onDragStart);
 
 let mLastDragOverTimestamp = null;
 
@@ -827,20 +867,26 @@ function isEventFiredOnTabDropBlocker(event) {
 }
 
 function onDragEnter(event) {
-  if (mDelayedDragEnter) {
-    clearTimeout(mDelayedDragEnter);
-    mDelayedDragEnter = null;
-  }
-  mDelayedDragEnter = setTimeout(() => {
-    mDelayedDragEnter = null;
-    mDraggingOnSelfWindow = true;
-    if (mDelayedDragLeave) {
-      clearTimeout(mDelayedDragLeave);
-      mDelayedDragLeave = null;
-    }
-  }, 10);
+  mDraggingOnSelfWindow = true;
 
   const info = getDropAction(event);
+  try {
+    const enteredTab = EventUtils.getTabFromEvent(event);
+    const leftTab    = Tabs.getTabFromChild(event.relatedTarget);
+    if (leftTab != enteredTab) {
+      mDraggingOnDraggedTabs = (
+        info.dragData &&
+        info.dragData.apiTabs.some(tab => tab.id == enteredTab.apiTab.id)
+      );
+    }
+    if (enteredTab.ownerDocument == document) {
+      Tabs.getTabsContainer(enteredTab).classList.add(kTABBAR_STATE_TAB_DRAGGING);
+      document.documentElement.classList.add(kTABBAR_STATE_TAB_DRAGGING);
+    }
+  }
+  catch(_e) {
+  }
+
   const dt   = event.dataTransfer;
   dt.dropEffect = info.dropEffect;
   if (info.dropEffect == 'link')
@@ -909,20 +955,49 @@ reserveToProcessLongHover.cancel = function() {
   clearTimeout(mLongHoverTimerNext);
 };
 
-function onDragLeave(_event) {
-  if (mDelayedDragLeave) {
-    clearTimeout(mDelayedDragLeave);
-    mDelayedDragLeave = null;
+function onDragLeave(event) {
+  let leftFromTabBar = false;
+  try {
+    const info       = getDropAction(event);
+    const leftTab    = EventUtils.getTabFromEvent(event);
+    const enteredTab = Tabs.getTabFromChild(event.relatedTarget);
+    if (leftTab != enteredTab) {
+      if (info.dragData &&
+          info.dragData.apiTabs.some(tab => tab.id == leftTab.apiTab.id) &&
+          (!enteredTab ||
+           !info.dragData.apiTabs.every(tab => tab.id == enteredTab.apiTab.id))) {
+        onDragLeave.delayedLeftFromDraggedTabs = setTimeout(() => {
+          delete onDragLeave.delayedLeftFromDraggedTabs;
+          mDraggingOnDraggedTabs = false;
+        }, 10);
+      }
+      else {
+        leftFromTabBar = !enteredTab || enteredTab.ownerDocument != document;
+        if (onDragLeave.delayedLeftFromDraggedTabs) {
+          clearTimeout(onDragLeave.delayedLeftFromDraggedTabs);
+          delete onDragLeave.delayedLeftFromDraggedTabs;
+        }
+      }
+    }
   }
-  setTimeout(() => {
-    mDelayedDragLeave = setTimeout(() => {
-      mDelayedDragLeave = null;
+  catch(_e) {
+    leftFromTabBar = true;
+  }
+
+  if (leftFromTabBar) {
+    onDragLeave.delayedLeftFromTabBar = setTimeout(() => {
+      delete onDragLeave.delayedLeftFromTabBar;
       mDraggingOnSelfWindow = false;
+      mDraggingOnDraggedTabs = false;
       clearDropPosition();
       clearDraggingState();
       mLastDropPosition = null;
-    }, configs.preventTearOffTabsTimeout);
-  }, 10);
+    }, 10);
+  }
+  else if (onDragLeave.delayedLeftFromTabBar) {
+    clearTimeout(onDragLeave.delayedLeftFromTabBar);
+    delete onDragLeave.delayedLeftFromTabBar;
+  }
 
   clearTimeout(mLongHoverTimer);
   mLongHoverTimer = null;
@@ -949,7 +1024,7 @@ function onDrop(event) {
   if (dropActionInfo.dragData &&
       dropActionInfo.dragData.apiTab) {
     log('there are dragged tabs');
-    Tree.performTabsDragDrop({
+    Commands.performTabsDragDrop({
       windowId:            dropActionInfo.dragData.windowId,
       tabs:                dropActionInfo.dragData.apiTabs,
       action:              dropActionInfo.action,
@@ -969,16 +1044,23 @@ function onDrop(event) {
 onDrop = EventUtils.wrapWithErrorHandler(onDrop);
 
 function onDragEnd(event) {
-  log('onDragEnd, mDraggingOnSelfWindow = ', mDraggingOnSelfWindow);
+  log('onDragEnd, ', { mDraggingOnSelfWindow, mDraggingOnDraggedTabs });
 
   let dragData = event.dataTransfer.getData(kTREE_DROP_TYPE);
   dragData = (dragData && JSON.parse(dragData)) || mCurrentDragData;
   if (Array.isArray(dragData.apiTabs))
     dragData.tabNodes = dragData.apiTabs.map(Tabs.getTabById);
 
-  finishDrag();
+  // Don't clear flags immediately, because they are referred by following operations in this function.
+  setTimeout(finishDrag, 0);
 
-  if (event.dataTransfer.dropEffect != 'none' ||
+  if (event.dataTransfer.getData(kTYPE_URI_LIST)) {
+    log('do nothing by TST for dropping just for bookmarking or linking');
+    return;
+  }
+
+  if (event.dataTransfer.mozUserCancelled ||
+      event.dataTransfer.dropEffect != 'none' ||
       //event.shiftKey || // don't ignore shift-drop, because it can be used to drag a parent tab as an individual tab.
       !configs.moveDroppedTabToNewWindowForUnhandledDragEvent) {
     log('dragged items are processed by someone: ', event.dataTransfer.dropEffect);
@@ -986,10 +1068,11 @@ function onDragEnd(event) {
   }
 
   const dropTargetTab = EventUtils.getTabFromEvent(event);
-  if (dropTargetTab &&
-      dragData &&
-      dragData.tabNodes &&
-      !dragData.tabNodes.includes(dropTargetTab)) {
+  if (mDraggingOnDraggedTabs ||
+      (dropTargetTab &&
+       dragData &&
+       dragData.tabNodes &&
+       !dragData.tabNodes.includes(dropTargetTab))) {
     log('ignore drop on dragged tabs themselves');
     return;
   }
@@ -1002,8 +1085,10 @@ function onDragEnd(event) {
     windowY,
     windowW: window.innerWidth,
     windowH: window.innerHeight,
-    eventX:  event.screenX,
-    eventY:  event.screenY,
+    eventScreenX: event.screenX,
+    eventScreenY: event.screenY,
+    eventClientX: event.clientX,
+    eventClientY: event.clientY,
     offset
   });
   if (event.screenX >= windowX - offset &&
@@ -1018,7 +1103,7 @@ function onDragEnd(event) {
   event.stopPropagation();
   event.preventDefault();
 
-  if (isDraggingAllCurrentTabs(dragData.tabNode)) {
+  if (isDraggingAllActiveTabs(dragData.tabNode)) {
     log('all tabs are dragged, so it is nonsence to tear off them from the window');
     return;
   }
@@ -1036,6 +1121,12 @@ function finishDrag() {
   log('finishDrag');
   clearDraggingTabsState();
 
+  mDragBehaviorNotification.classList.add('hiding');
+  mDragBehaviorNotification.classList.remove('shown');
+  setTimeout(() => {
+    mDragBehaviorNotification.classList.remove('hiding');
+  }, configs.collapseDuration);
+
   mDraggingOnSelfWindow = false;
 
   wait(100).then(() => {
@@ -1052,6 +1143,8 @@ function finishDrag() {
   mLastDragOverTimestamp = null;
   clearDraggingState();
   collapseAutoExpandedTabsWhileDragging();
+  mDraggingOnSelfWindow = false;
+  mDraggingOnDraggedTabs = false;
 }
 
 
@@ -1068,10 +1161,12 @@ function onTSTAPIDragEnter(event) {
       (!mDragTargetIsClosebox ||
        EventUtils.isEventFiredOnClosebox(event))) {
     if (target != mLastDragEnteredTarget) {
+      const windowId = Tabs.getWindow();
       TSTAPI.sendMessage({
         type:   TSTAPI.kNOTIFY_TAB_DRAGENTER,
         tab:    TSTAPI.serializeTab(tab),
-        window: Tabs.getWindow()
+        window: windowId,
+        windowId
       });
     }
   }
@@ -1091,10 +1186,12 @@ function onTSTAPIDragExit(event) {
   cancelDelayedTSTAPIDragExitOn(target);
   target.onTSTAPIDragExitTimeout = setTimeout(() => {
     delete target.onTSTAPIDragExitTimeout;
+    const windowId = Tabs.getWindow();
     TSTAPI.sendMessage({
       type:   TSTAPI.kNOTIFY_TAB_DRAGEXIT,
       tab:    TSTAPI.serializeTab(tab),
-      window: Tabs.getWindow()
+      window: windowId,
+      windowId
     });
   }, 10);
 }
