@@ -5,8 +5,6 @@
 */
 'use strict';
 
-import TabFavIconHelper from '/extlib/TabFavIconHelper.js';
-
 import {
   log as internalLogger,
   dumpTab,
@@ -20,7 +18,6 @@ import * as Tabs from '/common/tabs.js';
 import * as TabsOpen from '/common/tabs-open.js';
 import * as TabsGroup from '/common/tabs-group.js';
 import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
-import * as TabsContainer from '/common/tabs-container.js';
 import * as Tree from '/common/tree.js';
 import * as TSTAPI from '/common/tst-api.js';
 
@@ -145,12 +142,11 @@ async function updateRelatedGroupTab(groupTab, changedInfo = []) {
       const opener = Tabs.getOpenerFromGroupTab(groupTab);
       if (opener) {
         if (opener &&
-            (opener.apiTab.favIconUrl ||
-             TabFavIconHelper.maybeImageTab(opener.apiTab))) {
+            opener.apiTab.favIconUrl) {
           browser.runtime.sendMessage({
             type:       Constants.kCOMMAND_NOTIFY_TAB_FAVICON_UPDATED,
             tab:        groupTab.id,
-            favIconUrl: opener.apiTab.favIconUrl || opener.apiTab.url
+            favIconUrl: opener.apiTab.favIconUrl
           });
         }
         newTitle = browser.i18n.getMessage('groupTab_fromPinnedTab_label', opener.apiTab.title);
@@ -175,67 +171,73 @@ Tabs.onRemoved.addListener(async (tab, _closeInfo = {}) => {
 });
 
 Tabs.onUpdated.addListener((tab, changeInfo) => {
-  const apiTab = tab && tab.apiTab && tab.apiTab;
-  const status = changeInfo.status || apiTab && apiTab.status;
-  const url = changeInfo.url ? changeInfo.url :
-    status == 'complete' && apiTab ? apiTab.url : '';
-  if (tab &&
-      apiTab &&
-      status == 'complete') {
-    if (url.indexOf(Constants.kGROUP_TAB_URI) == 0) {
-      tab.classList.add(Constants.kTAB_STATE_GROUP_TAB);
-      Tabs.addSpecialTabState(tab, Constants.kTAB_STATE_GROUP_TAB);
-    }
-    else if (!Constants.kSHORTHAND_ABOUT_URI.test(url)) {
-      Tabs.getSpecialTabState(tab).then(async (states) => {
-        if (url.indexOf(Constants.kGROUP_TAB_URI) == 0)
-          return;
-        // Detect group tab from different session - which can have different UUID for the URL.
-        const PREFIX_REMOVER = /^moz-extension:\/\/[^\/]+/;
-        const pathPart = url.replace(PREFIX_REMOVER, '');
-        if (states.includes(Constants.kTAB_STATE_GROUP_TAB) &&
-            pathPart.split('?')[0] == Constants.kGROUP_TAB_URI.replace(PREFIX_REMOVER, '')) {
-          const parameters = pathPart.replace(/^[^\?]+\?/, '');
-          const oldUrl = tab.apiTab.url;
-          await wait(100); // for safety
-          if (tab.apiTab.url != oldUrl)
+  if ('url' in changeInfo ||
+      'previousUrl' in changeInfo ||
+      'state' in changeInfo) {
+    const apiTab = tab && tab.apiTab && tab.apiTab;
+    const status = changeInfo.status || apiTab && apiTab.status;
+    const url = changeInfo.url ? changeInfo.url :
+      status == 'complete' && apiTab ? apiTab.url : '';
+    if (tab &&
+        apiTab &&
+        status == 'complete') {
+      if (url.indexOf(Constants.kGROUP_TAB_URI) == 0) {
+        tab.classList.add(Constants.kTAB_STATE_GROUP_TAB);
+        Tabs.addSpecialTabState(tab, Constants.kTAB_STATE_GROUP_TAB);
+      }
+      else if (!Constants.kSHORTHAND_ABOUT_URI.test(url)) {
+        Tabs.getSpecialTabState(tab).then(async (states) => {
+          if (url.indexOf(Constants.kGROUP_TAB_URI) == 0)
             return;
-          browser.tabs.update(tab.apiTab.id, {
-            url: `${Constants.kGROUP_TAB_URI}?${parameters}`
-          }).catch(ApiTabs.handleMissingTabError);
-          tab.classList.add(Constants.kTAB_STATE_GROUP_TAB);
-        }
-        else {
-          Tabs.removeSpecialTabState(tab, Constants.kTAB_STATE_GROUP_TAB);
-          tab.classList.remove(Constants.kTAB_STATE_GROUP_TAB);
-        }
+          // Detect group tab from different session - which can have different UUID for the URL.
+          const PREFIX_REMOVER = /^moz-extension:\/\/[^\/]+/;
+          const pathPart = url.replace(PREFIX_REMOVER, '');
+          if (states.includes(Constants.kTAB_STATE_GROUP_TAB) &&
+              pathPart.split('?')[0] == Constants.kGROUP_TAB_URI.replace(PREFIX_REMOVER, '')) {
+            const parameters = pathPart.replace(/^[^\?]+\?/, '');
+            const oldUrl = tab.apiTab.url;
+            await wait(100); // for safety
+            if (tab.apiTab.url != oldUrl)
+              return;
+            browser.tabs.update(tab.apiTab.id, {
+              url: `${Constants.kGROUP_TAB_URI}?${parameters}`
+            }).catch(ApiTabs.handleMissingTabError);
+            tab.classList.add(Constants.kTAB_STATE_GROUP_TAB);
+          }
+          else {
+            Tabs.removeSpecialTabState(tab, Constants.kTAB_STATE_GROUP_TAB);
+            tab.classList.remove(Constants.kTAB_STATE_GROUP_TAB);
+          }
+        });
+      }
+    }
+    // restored tab can be replaced with blank tab. we need to restore it manually.
+    else if (changeInfo.url == 'about:blank' &&
+             changeInfo.previousUrl &&
+             changeInfo.previousUrl.indexOf(Constants.kGROUP_TAB_URI) == 0) {
+      const oldUrl = apiTab.url;
+      wait(100).then(() => { // redirect with delay to avoid infinite loop of recursive redirections.
+        if (tab.apiTab.url != oldUrl)
+          return;
+        browser.tabs.update(tab.apiTab.id, {
+          url: changeInfo.previousUrl
+        }).catch(ApiTabs.handleMissingTabError);
+        tab.classList.add(Constants.kTAB_STATE_GROUP_TAB);
+        Tabs.addSpecialTabState(tab, Constants.kTAB_STATE_GROUP_TAB);
       });
     }
-  }
-  // restored tab can be replaced with blank tab. we need to restore it manually.
-  else if (changeInfo.url == 'about:blank' &&
-           changeInfo.previousUrl &&
-           changeInfo.previousUrl.indexOf(Constants.kGROUP_TAB_URI) == 0) {
-    const oldUrl = apiTab.url;
-    wait(100).then(() => { // redirect with delay to avoid infinite loop of recursive redirections.
-      if (tab.apiTab.url != oldUrl)
-        return;
-      browser.tabs.update(tab.apiTab.id, {
-        url: changeInfo.previousUrl
-      }).catch(ApiTabs.handleMissingTabError);
-      tab.classList.add(Constants.kTAB_STATE_GROUP_TAB);
-      Tabs.addSpecialTabState(tab, Constants.kTAB_STATE_GROUP_TAB);
-    });
+
+    if (changeInfo.status ||
+        changeInfo.url ||
+        url.indexOf(Constants.kGROUP_TAB_URI) == 0)
+      tryInitGroupTab(tab);
   }
 
-  if (changeInfo.status ||
-      changeInfo.url ||
-      url.indexOf(Constants.kGROUP_TAB_URI) == 0)
-    tryInitGroupTab(tab);
-
-  const group = Tabs.getGroupTabForOpener(tab);
-  if (group)
-    reserveToUpdateRelatedGroupTabs(group, ['title', 'tree']);
+  if ('title' in changeInfo) {
+    const group = Tabs.getGroupTabForOpener(tab);
+    if (group)
+      reserveToUpdateRelatedGroupTabs(group, ['title', 'tree']);
+  }
 });
 
 Tabs.onGroupTabDetected.addListener(tab => {
@@ -282,12 +284,12 @@ Tabs.onBeforeCreate.addListener(async (apiTab, info) => {
       (configs.autoGroupNewTabs &&
        !openerApiTab &&
        !info.maybeOrphan)) {
-    if (parseInt(container.dataset.preventAutoGroupNewTabsUntil) > Date.now()) {
-      TabsContainer.incrementCounter(container, 'preventAutoGroupNewTabsUntil', configs.autoGroupNewTabsTimeout);
+    if (container.preventAutoGroupNewTabsUntil > Date.now()) {
+      container.preventAutoGroupNewTabsUntil += configs.autoGroupNewTabsTimeout;
     }
     else {
-      container.dataset.openedNewTabs += `|${apiTab.id}`;
-      container.dataset.openedNewTabsOpeners += `|${openerApiTab && openerApiTab.id}`;
+      container.openedNewTabs.push(apiTab.id);
+      container.openedNewTabsOpeners.push(openerApiTab && openerApiTab.id);
     }
   }
   if (container.openedNewTabsTimeout)
@@ -302,23 +304,23 @@ Tabs.onBeforeCreate.addListener(async (apiTab, info) => {
 const mToBeGroupedTabSets = [];
 
 async function onNewTabsTimeout(container) {
-  if (Tabs.hasCreatingTab())
-    await Tabs.waitUntilAllTabsAreCreated();
-  if (Tabs.hasMovingTab())
-    await Tabs.waitUntilAllTabsAreMoved();
+  if (Tabs.hasCreatingTab(container.windowId))
+    await Tabs.waitUntilAllTabsAreCreated(container.windowId);
+  if (Tabs.hasMovingTab(container.windowId))
+    await Tabs.waitUntilAllTabsAreMoved(container.windowId);
 
-  const tabIds       = container.dataset.openedNewTabs.split('|').filter(part => part != '');
-  const tabOpenerIds = container.dataset.openedNewTabsOpeners.split('|').filter(part => part != '');
+  const tabIds       = container.openedNewTabs;
+  const tabOpenerIds = container.openedNewTabsOpeners;
   log('onNewTabsTimeout ', tabIds);
   let tabReferences = tabIds.map((id, index) => {
     return {
-      id:          parseInt(id),
-      openerTabId: parseInt(tabOpenerIds[index])
+      id,
+      openerTabId: tabOpenerIds[index]
     };
   });
 
-  container.dataset.openedNewTabs = '';
-  container.dataset.openedNewTabsOpeners = '';
+  container.openedNewTabs        = [];
+  container.openedNewTabsOpeners = [];
 
   tabReferences = tabReferences.filter(tabReference => tabReference.id != '');
   if (tabReferences.length == 0 ||
@@ -480,7 +482,7 @@ async function tryGroupNewTabsFromPinnedOpener(rootTabs) {
       // Prevent the tab to be grouped again after it is ungrouped manually.
       child.dataset.alreadyGroupedForPinnedOpener = true;
       await Tree.attachTabTo(child, parent, {
-        forceExpand: true, // this is required to avoid the group tab itself is focused from active tab in collapsed tree
+        forceExpand: true, // this is required to avoid the group tab itself is active from active tab in collapsed tree
         insertAfter: configs.insertNewChildAt == Constants.kINSERT_FIRST ? parent : Tabs.getLastDescendantTab(parent),
         broadcast: true
       });
