@@ -7,7 +7,8 @@
 
 import {
   log as internalLogger,
-  wait
+  wait,
+  configs
 } from '/common/common.js';
 
 import * as Constants from '/common/constants.js';
@@ -95,6 +96,8 @@ async function onShortcutCommand(command) {
     active:        true,
     currentWindow: true
   }))[0]);
+  const selectedTabs = Tabs.isMultiselected(activeTab) ? Tabs.getSelectedTabs(activeTab) : [];
+  log('onShortcutCommand ', { command, activeTab, selectedTabs });
 
   switch (command) {
     case '_execute_browser_action':
@@ -157,10 +160,16 @@ async function onShortcutCommand(command) {
       return;
 
     case 'newContainerTab':
-      return browser.runtime.sendMessage({
+      browser.runtime.sendMessage({
         type:     Constants.kCOMMAND_SHOW_CONTAINER_SELECTOR,
         windowId: activeTab.apiTab.windowId
       });
+      return;
+
+    case 'groupSelectedTabs':
+      if (selectedTabs.length > 1)
+        TabsGroup.groupTabs(selectedTabs, { broadcast: true });
+      return;
 
     case 'indent':
       Commands.indent(activeTab, { followChildren: true });
@@ -184,21 +193,21 @@ async function onShortcutCommand(command) {
 
     case 'focusPrevious':
     case 'focusPreviousSilently': {
-      const nextFocused = Tabs.getPreviousVisibleTab(activeTab) ||
+      const nextActive = Tabs.getPreviousVisibleTab(activeTab) ||
         Tabs.getLastVisibleTab(activeTab);
-      TabsInternalOperation.selectTab(nextFocused, { silently: /Silently/.test(command) });
+      TabsInternalOperation.activateTab(nextActive, { silently: /Silently/.test(command) });
     }; return;
     case 'focusNext':
     case 'focusNextSilently': {
-      const nextFocused = Tabs.getNextVisibleTab(activeTab) ||
+      const nextActive = Tabs.getNextVisibleTab(activeTab) ||
         Tabs.getFirstVisibleTab(activeTab);
-      TabsInternalOperation.selectTab(nextFocused, { silently: /Silently/.test(command) });
+      TabsInternalOperation.activateTab(nextActive, { silently: /Silently/.test(command) });
     }; return;
     case 'focusParent':
-      TabsInternalOperation.selectTab(Tabs.getParentTab(activeTab));
+      TabsInternalOperation.activateTab(Tabs.getParentTab(activeTab));
       return;
     case 'focusFirstChild':
-      TabsInternalOperation.selectTab(Tabs.getFirstChildTab(activeTab));
+      TabsInternalOperation.activateTab(Tabs.getFirstChildTab(activeTab));
       return;
 
     case 'tabbarUp':
@@ -269,6 +278,9 @@ function onMessage(message, sender) {
         });
       })();
 
+    case Constants.kCOMMAND_PULL_TABS_ORDER:
+      return Promise.resolve(Tabs.getAllTabs(message.windowId).map(tab => tab.apiTab.id));
+
     case Constants.kCOMMAND_PULL_TREE_STRUCTURE:
       return (async () => {
         while (!mInitialized) {
@@ -362,9 +374,8 @@ function onMessage(message, sender) {
         logMouseEvent('Sending message to listeners');
         const serializedTab = TSTAPI.serializeTab(tab);
         const mousedownNotified = TSTAPI.sendMessage(Object.assign({}, message, {
-          type:   TSTAPI.kNOTIFY_TAB_MOUSEDOWN,
-          tab:    serializedTab,
-          window: tab.apiTab.windowId
+          type: TSTAPI.kNOTIFY_TAB_MOUSEDOWN,
+          tab:  serializedTab
         }));
 
         // We must send tab-mouseup after tab-mousedown is notified.
@@ -372,21 +383,35 @@ function onMessage(message, sender) {
         mousedownNotified.then(async (results) => {
           results = results.concat(
             await TSTAPI.sendMessage(Object.assign({}, message, {
-              type:   TSTAPI.kNOTIFY_TAB_CLICKED,
-              tab:    serializedTab,
-              window: tab.apiTab.windowId
+              type: TSTAPI.kNOTIFY_TAB_CLICKED,
+              tab:  serializedTab
             }))
           );
-          if (results.some(result => result.result)) // canceled
-            return;
+          if (results.some(result => result.result))
+            return browser.runtime.sendMessage({
+              type:     Constants.kNOTIFY_TAB_MOUSEDOWN_CANCELED,
+              windowId: message.windowId,
+              button:   message.button
+            });
 
-          logMouseEvent('Ready to select the tab');
+          logMouseEvent('Ready to handle click action on the tab');
 
           // not canceled, then fallback to default behavior
-          const wasMultiselectionAction = await HandleTabMultiselect.updateSelectionByTabClick(tab, message);
+          const onRegularArea = (
+            !message.twisty &&
+            !message.soundButton &&
+            !message.closebox
+          );
+          const wasMultiselectionAction = (
+            onRegularArea &&
+            await HandleTabMultiselect.updateSelectionByTabClick(tab, message)
+          );
           if (message.button == 0 &&
+              onRegularArea &&
               !wasMultiselectionAction)
-            TabsInternalOperation.selectTab(tab);
+            TabsInternalOperation.activateTab(tab, {
+              keepMultiselection: tab.apiTab.highlighted
+            });
         });
 
         return true;
@@ -408,7 +433,7 @@ function onMessage(message, sender) {
         const tab = Tabs.getTabById(message.tab);
         if (!tab)
           return;
-        TabsInternalOperation.selectTab(tab, Object.assign({}, message.options, {
+        TabsInternalOperation.activateTab(tab, Object.assign({}, message.options, {
           inRemote: false
         }));
       })();
@@ -524,7 +549,7 @@ function onMessage(message, sender) {
           message.insertAfter
         ]);
         log('perform tabs dragdrop requested: ', message);
-        return Tree.performTabsDragDrop(Object.assign({}, message, {
+        return Commands.performTabsDragDrop(Object.assign({}, message, {
           attachTo:     Tabs.getTabById(message.attachTo),
           insertBefore: Tabs.getTabById(message.insertBefore),
           insertAfter:  Tabs.getTabById(message.insertAfter)
@@ -558,7 +583,7 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kGET_TREE:
       return (async () => {
         const tabs    = await TSTAPI.getTargetTabs(message, sender);
-        const results = tabs.map(TSTAPI.serializeTab);
+        const results = await Promise.all(tabs.map(TSTAPI.serializeTabWithEffectiveFavIconUrl));
         return TSTAPI.formatResult(results, message);
       })();
 
@@ -643,6 +668,13 @@ function onMessageExternal(message, sender) {
         return TSTAPI.formatResult(results, message);
       })();
 
+    case TSTAPI.kMOVE_TO_START:
+      return (async () => {
+        const tabs = await TSTAPI.getTargetTabs(message, sender);
+        await Commands.moveTabsToStart(tabs);
+        return true;
+      })();
+
     case TSTAPI.kMOVE_DOWN:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
@@ -650,11 +682,18 @@ function onMessageExternal(message, sender) {
         return TSTAPI.formatResult(results, message);
       })();
 
+    case TSTAPI.kMOVE_TO_END:
+      return (async () => {
+        const tabs = await TSTAPI.getTargetTabs(message, sender);
+        await Commands.moveTabsToEnd(tabs);
+        return true;
+      })();
+
     case TSTAPI.kFOCUS:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
         for (const tab of tabs) {
-          TabsInternalOperation.selectTab(tab, {
+          TabsInternalOperation.activateTab(tab, {
             silently: message.silently
           });
         }
@@ -664,7 +703,7 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kDUPLICATE:
       return (async () => {
         const tabs   = await TSTAPI.getTargetTabs(message, sender);
-        let behavior = Constants.kNEWTAB_OPEN_AS_ORPHAN;
+        let behavior = configs.autoAttachOnDuplicated;
         switch (String(message.as || 'sibling').toLowerCase()) {
           case 'child':
             behavior = Constants.kNEWTAB_OPEN_AS_CHILD;
@@ -675,19 +714,17 @@ function onMessageExternal(message, sender) {
           case 'nextsibling':
             behavior = Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING;
             break;
+          case 'orphan':
+            behavior = Constants.kNEWTAB_OPEN_AS_ORPHAN;
+            break;
           default:
             break;
         }
         for (const tab of tabs) {
-          const duplicatedTabs = await Tree.moveTabs([tab], {
-            duplicate:           true,
+          await Commands.duplicateTab(tab, {
             destinationWindowId: tab.apiTab.windowId,
-            insertAfter:         tab
-          });
-          await Tree.behaveAutoAttachedTab(duplicatedTabs[0], {
-            broadcast: true,
-            baseTab:   tab,
-            behavior
+            behavior,
+            multiselected: false
           });
         }
         return TSTAPI.formatResult(tabs.map(() => true), message);
@@ -700,6 +737,22 @@ function onMessageExternal(message, sender) {
         return groupTab.apiTab;
       })();
 
+    case TSTAPI.kOPEN_IN_NEW_WINDOW:
+      return (async () => {
+        const tabs = await TSTAPI.getTargetTabs(message, sender);
+        const windowId = await Commands.openTabsInWindow(tabs, {
+          multiselected: false
+        });
+        return windowId;
+      })();
+
+    case TSTAPI.kREOPEN_IN_CONTAINER:
+      return (async () => {
+        const tabs = await TSTAPI.getTargetTabs(message, sender);
+        const reopenedTabs = await Commands.reopenInContainer(tabs, message.containerId || 'firefox-default');
+        return TSTAPI.formatResult(reopenedTabs.map(tab => tab.apiTab), message);
+      })();
+
     case TSTAPI.kGET_TREE_STRUCTURE:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
@@ -748,5 +801,22 @@ function onMessageExternal(message, sender) {
         });
         return true;
       })();
+
+    case TSTAPI.kGRANT_TO_REMOVE_TABS:
+      return (async () => {
+        const tabs = await TSTAPI.getTargetTabs(message, sender);
+        const grantedRemovingTabIds = configs.grantedRemovingTabIds.concat(tabs.filter(Tabs.ensureLivingTab).map(tab => tab.apiTab.id));
+        configs.grantedRemovingTabIds = Array.from(new Set(grantedRemovingTabIds));
+        return true;
+      })();
+
+    case TSTAPI.kSTART_CUSTOM_DRAG:
+      return (async () => {
+        return browser.runtime.sendMessage({
+          type:     Constants.kNOTIFY_TAB_MOUSEDOWN_EXPIRED,
+          windowId: message.windowId || (await browser.windows.getLastFocused({ populate: false })).id,
+          button:   message.button || 0
+        });
+      })();
   }
 }
