@@ -3,8 +3,8 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
-import '../extlib/l10n.js';
-import '../extlib/TabIdFixer.js';
+import '/extlib/l10n.js';
+import '/extlib/TabIdFixer.js';
 
 import {
   log
