@@ -152,6 +152,7 @@ function onMouseMove(event) {
       type:     TSTAPI.kNOTIFY_TAB_MOUSEMOVE,
       tab:      TSTAPI.serializeTab(tab),
       window:   mTargetWindow,
+      windowId: mTargetWindow,
       ctrlKey:  event.ctrlKey,
       shiftKey: event.shiftKey,
       altKey:   event.altKey,
@@ -177,6 +178,7 @@ function onMouseOver(event) {
       type:     TSTAPI.kNOTIFY_TAB_MOUSEOVER,
       tab:      TSTAPI.serializeTab(tab),
       window:   mTargetWindow,
+      windowId: mTargetWindow,
       ctrlKey:  event.ctrlKey,
       shiftKey: event.shiftKey,
       altKey:   event.altKey,
@@ -202,6 +204,7 @@ function onMouseOut(event) {
       type:     TSTAPI.kNOTIFY_TAB_MOUSEOUT,
       tab:      TSTAPI.serializeTab(tab),
       window:   mTargetWindow,
+      windowId: mTargetWindow,
       ctrlKey:  event.ctrlKey,
       shiftKey: event.shiftKey,
       altKey:   event.altKey,
@@ -238,6 +241,7 @@ function onMouseDown(event) {
   const mousedownDetail = {
     targetType:    getMouseEventTargetType(event),
     tab:           tab && tab.id,
+    twisty:        EventUtils.isEventFiredOnTwisty(event),
     soundButton:   EventUtils.isEventFiredOnSoundButton(event),
     closebox:      EventUtils.isEventFiredOnClosebox(event),
     button:        event.button,
@@ -264,17 +268,14 @@ function onMouseDown(event) {
     promisedMousedownNotified: Promise.resolve()
   };
 
-  if ((!EventUtils.isEventFiredOnTwisty(event) &&
-       !EventUtils.isEventFiredOnSoundButton(event) &&
-       !EventUtils.isEventFiredOnClosebox(event)) ||
-      event.button != 0)
-    mousedown.promisedMousedownNotified = browser.runtime.sendMessage(Object.assign({}, mousedownDetail, {
-      type:     Constants.kNOTIFY_TAB_MOUSEDOWN,
-      windowId: mTargetWindow
-    }));
+  mousedown.promisedMousedownNotified = browser.runtime.sendMessage(Object.assign({}, mousedownDetail, {
+    type:     Constants.kNOTIFY_TAB_MOUSEDOWN,
+    window:   mTargetWindow,
+    windowId: mTargetWindow,
+  }));
 
   EventUtils.setLastMousedown(event.button, mousedown);
-  mousedown.timeout = setTimeout(() => {
+  mousedown.timeout = setTimeout(async () => {
     if (!EventUtils.getLastMousedown(event.button))
       return;
 
@@ -291,17 +292,24 @@ function onMouseDown(event) {
       return;
     }
 
-    if (TSTAPI.getListenersForMessageType(TSTAPI.kNOTIFY_TAB_DRAGREADY).length == 0)
+    if (TSTAPI.getListenersForMessageType(TSTAPI.kNOTIFY_TAB_DRAGREADY).length == 0 ||
+        EventUtils.getElementTarget(event).closest('#tab-drag-handle'))
       return;
 
-    log('onMouseDown expired');
-    mousedown.expired = true;
-    if (event.button == 0) {
-      if (tab) {
-        DragAndDrop.startMultiDrag(tab, mousedown.detail.closebox);
+    if (event.button == 0 &&
+        tab) {
+      const results = await DragAndDrop.legacyStartMultiDrag(tab, mousedown.detail.closebox);
+      if (results.some(result => result.result !== false)) {
+        log('onMouseDown expired');
+        mousedown.expired = true;
+        onMessage({
+          type:     Constants.kNOTIFY_TAB_MOUSEDOWN_EXPIRED,
+          windowId: mTargetWindow,
+          button:   event.button
+        });
       }
     }
-  }, configs.startDragTimeout);
+  }, configs.longPressDuration);
 }
 onMouseDown = EventUtils.wrapWithErrorHandler(onMouseDown);
 
@@ -334,6 +342,12 @@ async function onMouseUp(event) {
   const livingTab = EventUtils.getTabFromEvent(event);
   log('onMouseUp: ', tab, { living: !!livingTab });
 
+  DragAndDrop.endMultiDrag(livingTab, event);
+
+  if (EventUtils.isEventFiredOnMenuOrPanel(event) ||
+      EventUtils.isEventFiredOnAnchor(event))
+    return;
+
   const lastMousedown = EventUtils.getLastMousedown(event.button);
   EventUtils.cancelHandleMousedown(event.button);
   if (lastMousedown)
@@ -345,16 +359,16 @@ async function onMouseUp(event) {
     const results = TSTAPI.sendMessage(Object.assign({}, lastMousedown.detail, {
       type:    TSTAPI.kNOTIFY_TAB_MOUSEUP,
       tab:     serializedTab,
-      window:  mTargetWindow
+      window:  mTargetWindow,
+      windowId: mTargetWindow
     }));
     // don't wait here, because we need process following common operations
     // even if this mouseup event is canceled.
     promisedCanceled = results.then(results => results.some(result => result.result));
   }
 
-  DragAndDrop.endMultiDrag(livingTab, event);
-
   if (!lastMousedown ||
+      lastMousedown.expired ||
       lastMousedown.detail.targetType != getMouseEventTargetType(event) ||
       (livingTab && livingTab != Tabs.getTabById(lastMousedown.detail.tab)))
     return;
@@ -370,28 +384,80 @@ async function onMouseUp(event) {
     if (lastMousedown.detail.isMiddleClick) { // Ctrl-click doesn't close tab on Firefox's tab bar!
       log('onMouseUp: middle click on a tab');
       //log('middle-click to close');
-      Sidebar.confirmToCloseTabs(Tree.getClosingTabsFromParent(livingTab).length)
-        .then(aConfirmed => {
-          if (aConfirmed)
+      const tabs = Tree.getClosingTabsFromParent(livingTab);
+      Sidebar.confirmToCloseTabs(tabs.map(tab => tab.apiTab.id))
+        .then(confirmed => {
+          if (confirmed)
             TabsInternalOperation.removeTab(livingTab, { inRemote: true });
         });
     }
+    else if (lastMousedown.detail.twisty) {
+      log('clicked on twisty');
+      if (Tabs.hasChildTabs(tab))
+        Tree.collapseExpandSubtree(tab, {
+          collapsed:       !Tabs.isSubtreeCollapsed(tab),
+          manualOperation: true,
+          inRemote:        true
+        });
+    }
+    else if (lastMousedown.detail.soundButton) {
+      log('clicked on sound button');
+      browser.runtime.sendMessage({
+        type:     Constants.kCOMMAND_SET_SUBTREE_MUTED,
+        windowId: mTargetWindow,
+        tab:      tab.id,
+        muted:    Tabs.maybeSoundPlaying(tab)
+      });
+    }
+    else if (lastMousedown.detail.closebox) {
+      log('clicked on closebox');
+      //if (!warnAboutClosingTabSubtreeOf(tab)) {
+      //  event.stopPropagation();
+      //  event.preventDefault();
+      //  return;
+      //}
+      const multiselected  = Tabs.isMultiselected(tab);
+      const tabsToBeClosed = multiselected ?
+        Tabs.getSelectedTabs(tab) :
+        Tree.getClosingTabsFromParent(tab) ;
+      Sidebar.confirmToCloseTabs(tabsToBeClosed.map(tab => tab.apiTab.id))
+        .then(confirmed => {
+          if (!confirmed)
+            return;
+          if (multiselected)
+            TabsInternalOperation.removeTabs(tabsToBeClosed, { inRemote: true });
+          else
+            TabsInternalOperation.removeTab(tab, { inRemote: true });
+        });
+    }
+    else if (lastMousedown.detail.button == 0 &&
+             !lastMousedown.detail.altKey &&
+             !lastMousedown.detail.ctrlKey &&
+             !lastMousedown.detail.metaKey &&
+             !lastMousedown.detail.shiftKey) {
+      // clear selection by left click
+      browser.tabs.highlight({
+        windowId: tab.apiTab.windowId,
+        tabs:     [tab.apiTab.index],
+        populate: false
+      }).catch(_e => {});
+    }
     return;
   }
-
-  if (tab) // ignore mouseup on closing tab or something
+  else if (tab) // ignore mouseup on closing tab or something
     return;
 
   // following codes are for handlig of click event on the tab bar itself.
   const actionForNewTabCommand = lastMousedown.detail.isAccelClick ?
     configs.autoAttachOnNewTabButtonMiddleClick :
     configs.autoAttachOnNewTabCommand;
-  if (EventUtils.isEventFiredOnNewTabButton(event) &&
-      lastMousedown.detail.button != 2) {
-    log('onMouseUp: click on the new tab button');
-    handleNewTabAction(event, {
-      action: actionForNewTabCommand
-    });
+  if (EventUtils.isEventFiredOnNewTabButton(event)) {
+    if (lastMousedown.detail.button != 2) {
+      log('onMouseUp: click on the new tab button');
+      handleNewTabAction(event, {
+        action: actionForNewTabCommand
+      });
+    }
     return;
   }
 
@@ -399,10 +465,12 @@ async function onMouseUp(event) {
   let results = await TSTAPI.sendMessage(Object.assign({}, lastMousedown.detail, {
     type:   TSTAPI.kNOTIFY_TABBAR_MOUSEUP,
     window: mTargetWindow,
+    windowId: mTargetWindow
   }));
   results = results.concat(await TSTAPI.sendMessage(Object.assign({}, lastMousedown.detail, {
     type:   TSTAPI.kNOTIFY_TABBAR_CLICKED,
     window: mTargetWindow,
+    windowId: mTargetWindow
   })));
   if (results.some(result => result.result))// canceled
     return;
@@ -416,91 +484,21 @@ async function onMouseUp(event) {
 }
 onMouseUp = EventUtils.wrapWithErrorHandler(onMouseUp);
 
-function onClick(event) {
+function onClick(_event) {
   // clear unexpectedly left "dragging" state
   // (see also https://github.com/piroor/treestyletab/issues/1921 )
   DragAndDrop.clearDraggingTabsState();
-
-  if (event.button != 0) // ignore non-left click
-    return;
-
-  log('onClick', String(event.target));
-
-  if (EventUtils.isEventFiredOnMenuOrPanel(event) ||
-      EventUtils.isEventFiredOnAnchor(event))
-    return;
-
-  if (EventUtils.isEventFiredOnNewTabButton(event)) {
-    log('onClick: canceled / on newtab button');
-    event.stopPropagation();
-    event.preventDefault();
-    return;
-  }
-
-  const tab = EventUtils.getTabFromEvent(event);
-  log('clicked tab: ', tab);
-
-  if (EventUtils.isEventFiredOnTwisty(event)) {
-    event.stopPropagation();
-    event.preventDefault();
-    log('clicked on twisty');
-    if (Tabs.hasChildTabs(tab))
-      Tree.collapseExpandSubtree(tab, {
-        collapsed:       !Tabs.isSubtreeCollapsed(tab),
-        manualOperation: true,
-        inRemote:        true
-      });
-    return;
-  }
-
-  if (EventUtils.isEventFiredOnSoundButton(event)) {
-    event.stopPropagation();
-    event.preventDefault();
-    log('clicked on sound button');
-    browser.runtime.sendMessage({
-      type:     Constants.kCOMMAND_SET_SUBTREE_MUTED,
-      windowId: mTargetWindow,
-      tab:      tab.id,
-      muted:    Tabs.maybeSoundPlaying(tab)
-    });
-    return;
-  }
-
-  if (EventUtils.isEventFiredOnClosebox(event)) {
-    event.stopPropagation();
-    event.preventDefault();
-    log('clicked on closebox');
-    //if (!warnAboutClosingTabSubtreeOf(tab)) {
-    //  event.stopPropagation();
-    //  event.preventDefault();
-    //  return;
-    //}
-    const multiselected  = Tabs.isMultiselected(tab);
-    const tabsToBeClosed = multiselected ?
-      Tabs.getSelectedTabs(tab) :
-      Tree.getClosingTabsFromParent(tab) ;
-    Sidebar.confirmToCloseTabs(tabsToBeClosed.length)
-      .then(aConfirmed => {
-        if (!aConfirmed)
-          return;
-        if (multiselected)
-          TabsInternalOperation.removeTabs(tabsToBeClosed, { inRemote: true });
-        else
-          TabsInternalOperation.removeTab(tab, { inRemote: true });
-      });
-    return;
-  }
 }
 onClick = EventUtils.wrapWithErrorHandler(onClick);
 
 function handleNewTabAction(event, options = {}) {
-  log('handleNewTabAction');
+  log('handleNewTabAction ', { event, options });
 
   if (!configs.autoAttach && !('action' in options))
     options.action = Constants.kNEWTAB_DO_NOTHING;
 
   Commands.openNewTabAs({
-    baseTab:      Tabs.getCurrentTab(mTargetWindow),
+    baseTab:      Tabs.getActiveTab(mTargetWindow),
     as:           options.action,
     cookieStoreId: options.cookieStoreId,
     inBackground: event.shiftKey,
@@ -517,7 +515,17 @@ function onDblClick(event) {
   log('dblclick tab: ', tab, { living: !!livingTab });
 
   if (livingTab) {
-    if (configs.collapseExpandSubtreeByDblClick) {
+    if (configs.simulateCloseTabByDblclick &&
+        event.button == 0 &&
+        !event.altKey &&
+        !event.ctrlKey &&
+        !event.metaKey &&
+        !event.shiftKey) {
+      event.stopPropagation();
+      event.preventDefault();
+      TabsInternalOperation.removeTab(livingTab, { inRemote: true });
+    }
+    else if (configs.collapseExpandSubtreeByDblClick) {
       event.stopPropagation();
       event.preventDefault();
       Tree.collapseExpandSubtree(livingTab, {
@@ -590,6 +598,27 @@ function onMessage(message, _sender, _respond) {
 
   //log('onMessage: ', message, sender);
   switch (message.type) {
+    case Constants.kNOTIFY_TAB_MOUSEDOWN_CANCELED:
+      if (message.windowId == mTargetWindow)
+        EventUtils.cancelHandleMousedown(message.button || 0);
+      break;
+
+    case Constants.kNOTIFY_TAB_MOUSEDOWN_EXPIRED:
+      if (message.windowId == mTargetWindow) {
+        const lastMousedown = EventUtils.getLastMousedown(message.button || 0);
+        if (lastMousedown)
+          lastMousedown.expired = true;
+      }
+      break;
+
+    case Constants.kCOMMAND_SHOW_CONTAINER_SELECTOR: {
+      const anchor = document.querySelector(`
+        :root.contextual-identity-selectable .contextual-identities-selector-anchor,
+        .newtab-button
+      `);
+      mContextualIdentitySelector.ui.open({ anchor });
+    }; break;
+
     case TSTAPI.kCOMMAND_BROADCAST_API_REGISTERED:
       wait(0).then(() => { // wait until addons are updated
         updateSpecialEventListenersForAPIListeners();
@@ -601,13 +630,5 @@ function onMessage(message, _sender, _respond) {
         updateSpecialEventListenersForAPIListeners();
       });
       break;
-
-    case Constants.kCOMMAND_SHOW_CONTAINER_SELECTOR: {
-      const anchor = document.querySelector(`
-        :root.contextual-identity-selectable .contextual-identities-selector-anchor,
-        .newtab-button
-      `);
-      mContextualIdentitySelector.ui.open({ anchor });
-    }; break;
   }
 }
