@@ -8,7 +8,6 @@
 import {
   log as internalLogger,
   wait,
-  nextFrame,
   configs
 } from '/common/common.js';
 
@@ -16,9 +15,9 @@ import * as Constants from '/common/constants.js';
 import * as Tabs from '/common/tabs.js';
 import * as TabsUpdate from '/common/tabs-update.js';
 import * as Tree from '/common/tree.js';
+import { SequenceMatcher } from '/common/diff.js';
 import TabFavIconHelper from '/extlib/TabFavIconHelper.js';
 
-// eslint-disable-next-line no-unused-vars
 function log(...args) {
   internalLogger('sidebar/sidebar-tabs', ...args);
 }
@@ -27,7 +26,7 @@ let mInitialized = false;
 
 export function init() {
   mInitialized = true;
-  document.querySelector('#master-throbber').addEventListener('animationiteration', synchronizeThrobberAnimation);
+  document.querySelector('#sync-throbber').addEventListener('animationiteration', synchronizeThrobberAnimation);
 
   const tabbar = document.querySelector('#tabbar');
   tabbar.addEventListener('overflow', onOverflow);
@@ -55,7 +54,17 @@ export function getClosebox(tab) {
 }
 
 
-function updateTwisty(tab) {
+export function reserveToUpdateTwistyTooltip(tab) {
+  if (tab.reservedUpdateTwistyTooltip)
+    return;
+  tab.reservedUpdateTwistyTooltip = () => {
+    delete tab.reservedUpdateTwistyTooltip;
+    updateTwistyTooltip(tab);
+  };
+  tab.addEventListener('mouseover', tab.reservedUpdateTwistyTooltip, { once: true });
+}
+
+function updateTwistyTooltip(tab) {
   let tooltip;
   if (Tabs.isSubtreeCollapsed(tab))
     tooltip = browser.i18n.getMessage('tab_twisty_collapsed_tooltip');
@@ -64,7 +73,17 @@ function updateTwisty(tab) {
   getTwisty(tab).setAttribute('title', tooltip);
 }
 
-function updateClosebox(tab) {
+export function reserveToUpdateCloseboxTooltip(tab) {
+  if (tab.reservedUpdateCloseboxTooltip)
+    return;
+  tab.reservedUpdateCloseboxTooltip = () => {
+    delete tab.reservedUpdateCloseboxTooltip;
+    updateCloseboxTooltip(tab);
+  };
+  tab.addEventListener('mouseover', tab.reservedUpdateCloseboxTooltip, { once: true });
+}
+
+function updateCloseboxTooltip(tab) {
   let tooltip;
   if (Tabs.isMultiselected(tab))
     tooltip = browser.i18n.getMessage('tab_closebox_tab_tooltip_multiselected');
@@ -86,19 +105,56 @@ function updateDescendantsCount(tab) {
   counter.textContent = count;
 }
 
-
-function reserveToUpdateTooltip(tab) {
-  if (!mInitialized ||
-      !Tabs.ensureLivingTab(tab))
+function updateDescendantsHighlighted(tab) {
+  const children = Tabs.getChildTabs(tab);
+  if (!Tabs.hasChildTabs(tab)) {
+    tab.classList.remove(Constants.kTAB_STATE_SOME_DESCENDANTS_HIGHLIGHTED);
+    tab.classList.remove(Constants.kTAB_STATE_ALL_DESCENDANTS_HIGHLIGHTED);
     return;
-  for (const updateTab of [tab].concat(Tabs.getAncestorTabs(tab))) {
-    if (updateTab.reservedUpdateTabTooltip)
-      clearTimeout(updateTab.reservedUpdateTabTooltip);
   }
-  tab.reservedUpdateTabTooltip = setTimeout(() => {
+  let someHighlighted = false;
+  let allHighlighted  = true;
+  for (const child of children) {
+    if (child.classList.contains(Constants.kTAB_STATE_HIGHLIGHTED)) {
+      someHighlighted = true;
+      allHighlighted = (
+        allHighlighted &&
+        (!Tabs.hasChildTabs(child) ||
+         child.classList.contains(Constants.kTAB_STATE_ALL_DESCENDANTS_HIGHLIGHTED))
+      );
+    }
+    else {
+      if (!someHighlighted &&
+          child.classList.contains(Constants.kTAB_STATE_SOME_DESCENDANTS_HIGHLIGHTED)) {
+        someHighlighted = true;
+      }
+      allHighlighted = false;
+    }
+  }
+  if (someHighlighted) {
+    tab.classList.add(Constants.kTAB_STATE_SOME_DESCENDANTS_HIGHLIGHTED);
+    if (allHighlighted)
+      tab.classList.add(Constants.kTAB_STATE_ALL_DESCENDANTS_HIGHLIGHTED);
+    else
+      tab.classList.remove(Constants.kTAB_STATE_ALL_DESCENDANTS_HIGHLIGHTED);
+  }
+  else {
+    tab.classList.remove(Constants.kTAB_STATE_SOME_DESCENDANTS_HIGHLIGHTED);
+    tab.classList.remove(Constants.kTAB_STATE_ALL_DESCENDANTS_HIGHLIGHTED);
+  }
+}
+
+
+export function reserveToUpdateTooltip(tab) {
+  if (!mInitialized ||
+      !Tabs.ensureLivingTab(tab) ||
+      tab.reservedUpdateTabTooltip)
+    return;
+  tab.reservedUpdateTabTooltip = () => {
     delete tab.reservedUpdateTabTooltip;
     updateTabAndAncestorsTooltip(tab);
-  }, 100);
+  };
+  tab.addEventListener('mouseover', tab.reservedUpdateTabTooltip, { once: true });
 }
 
 function updateTabAndAncestorsTooltip(tab) {
@@ -115,6 +171,32 @@ function updateTooltip(tab) {
 
   tab.dataset.labelWithDescendants = Tabs.getLabelWithDescendants(tab);
 
+  if (configs.debug) {
+    tab.dataset.label = `
+${tab.apiTab.title}
+#${tab.id}
+(${tab.className})
+uniqueId = <%${Constants.kPERSISTENT_ID}%>
+duplicated = <%duplicated%> / <%originalTabId%> / <%originalId%>
+restored = <%restored%>
+tabId = ${tab.apiTab.id}
+windowId = ${tab.apiTab.windowId}
+`.trim();
+    tab.setAttribute('title', tab.dataset.label);
+    tab.uniqueId.then(uniqueId => {
+      if (!Tabs.ensureLivingTab(tab))
+        return;
+      tab.setAttribute('title',
+                       tab.dataset.label = tab.dataset.label
+                         .replace(`<%${Constants.kPERSISTENT_ID}%>`, uniqueId.id)
+                         .replace(`<%originalId%>`, uniqueId.originalId)
+                         .replace(`<%originalTabId%>`, uniqueId.originalTabId)
+                         .replace(`<%duplicated%>`, !!uniqueId.duplicated)
+                         .replace(`<%restored%>`, !!uniqueId.restored));
+    });
+    return;
+  }
+
   if (configs.showCollapsedDescendantsByTooltip &&
       Tabs.isSubtreeCollapsed(tab) &&
       Tabs.hasChildTabs(tab)) {
@@ -122,9 +204,6 @@ function updateTooltip(tab) {
     return;
   }
 
-  if (configs.debug)
-    return;
-
   const label = Tabs.getTabLabel(tab);
   if (Tabs.isPinned(tab) || label.classList.contains('overflow')) {
     tab.setAttribute('title', tab.dataset.label);
@@ -158,17 +237,27 @@ async function synchronizeThrobberAnimation() {
   if (toBeSynchronizedTabs.length == 0)
     return;
 
-  for (const tab of Array.slice(toBeSynchronizedTabs)) {
+  for (const tab of toBeSynchronizedTabs) {
     tab.classList.remove(Constants.kTAB_STATE_THROBBER_UNSYNCHRONIZED);
   }
-  await nextFrame();
+
   document.documentElement.classList.add(Constants.kTABBAR_STATE_THROBBER_SYNCHRONIZING);
-  await nextFrame();
+  void document.documentElement.offsetWidth;
   document.documentElement.classList.remove(Constants.kTABBAR_STATE_THROBBER_SYNCHRONIZING);
 }
 
 
-export function updateSoundButtonTooltip(tab) {
+export function reserveToUpdateSoundButtonTooltip(tab) {
+  if (tab.reservedUpdateSoundButtonTooltip)
+    return;
+  tab.reservedUpdateSoundButtonTooltip = () => {
+    delete tab.reservedUpdateSoundButtonTooltip;
+    updateSoundButtonTooltip(tab);
+  };
+  tab.addEventListener('mouseover', tab.reservedUpdateSoundButtonTooltip, { once: true });
+}
+
+function updateSoundButtonTooltip(tab) {
   let tooltip = '';
   const suffix = Tabs.isMultiselected(tab) ? '_multiselected' : '' ;
   if (Tabs.maybeMuted(tab))
@@ -183,14 +272,27 @@ export function updateSoundButtonTooltip(tab) {
 export function updateAll() {
   updateLoadingState();
   synchronizeThrobberAnimation();
-  for (const tab of Tabs.getAllTabs()) {
-    updateTwisty(tab);
-    updateClosebox(tab);
+  // We need to update from bottom to top, because
+  // updateDescendantsHighlighted() refers results of descendants.
+  for (const tab of Tabs.getAllTabs().reverse()) {
+    reserveToUpdateTwistyTooltip(tab);
+    reserveToUpdateCloseboxTooltip(tab);
     updateDescendantsCount(tab);
+    updateDescendantsHighlighted(tab);
     updateTooltip(tab);
+    if (!Tabs.isCollapsed(tab))
+      updateLabelOverflow(tab);
   }
 }
 
+export function updateLabelOverflow(tab) {
+  const label = Tabs.getTabLabel(tab);
+  if (!Tabs.isPinned(tab) &&
+      label.firstChild.getBoundingClientRect().width > label.getBoundingClientRect().width)
+    label.classList.add('overflow');
+  else
+    label.classList.remove('overflow');
+}
 
 function onOverflow(event) {
   const tab = Tabs.getTabFromChild(event.target);
@@ -210,6 +312,103 @@ function onUnderflow(event) {
   }
 }
 
+
+
+export function reserveToSyncTabsOrder() {
+  if (reserveToSyncTabsOrder.timer)
+    clearTimeout(reserveToSyncTabsOrder.timer);
+  reserveToSyncTabsOrder.timer = setTimeout(() => {
+    delete reserveToSyncTabsOrder.timer;
+    syncTabsOrder();
+  }, 100);
+}
+reserveToSyncTabsOrder.retryCount = 0;
+
+async function syncTabsOrder() {
+  log('syncTabsOrder');
+  const windowId  = Tabs.getWindow();
+  const apiTabIds = await browser.runtime.sendMessage({
+    type: Constants.kCOMMAND_PULL_TABS_ORDER,
+    windowId
+  });
+  log('syncTabsOrder: apiTabIds = ', apiTabIds);
+  const container = Tabs.getTabsContainer();
+  if (container.childNodes.length != apiTabIds.length) {
+    if (reserveToSyncTabsOrder.retryCount > 10)
+      throw new Error(`fatal error: mismatched number of tabs in the window ${windowId}`);
+    log('syncTabsOrder: retry');
+    reserveToSyncTabsOrder.retryCount++;
+    return reserveToSyncTabsOrder();
+  }
+  reserveToSyncTabsOrder.retryCount = 0;
+
+
+  const currentApiTabIds = Array.from(container.childNodes, tab => tab.apiTab.id);
+  const DOMElementsOperations = (new SequenceMatcher(currentApiTabIds, apiTabIds)).operations();
+  const movedTabs = new Set();
+  const needToBeReindexedTabs = new Set();
+  for (const operation of DOMElementsOperations) {
+    const [tag, fromStart, fromEnd, toStart, toEnd] = operation;
+    log('syncTabsOrder: operation ', { tag, fromStart, fromEnd, toStart, toEnd });
+    switch (tag) {
+      case 'equal':
+      case 'delete':
+        break;
+
+      case 'insert':
+      case 'replace':
+        const moveTabIds = apiTabIds.slice(toStart, toEnd);
+        const referenceId = currentApiTabIds[fromStart] || null;
+        log(`syncTabsOrder: move ${moveTabIds.join(',')} before ${referenceId}`);
+        const referenceTab = referenceId && Tabs.getTabById(referenceId);
+        let lastNextOfMovedTab = null;
+        for (let i = 0, maxi = moveTabIds.length; i < maxi; i++) {
+          const id = moveTabIds[i];
+          const tab = Tabs.getTabById(id);
+          if (!tab)
+            continue;
+          if (i == maxi - 1)
+            lastNextOfMovedTab = Tabs.getNextTab(tab);
+          container.insertBefore(tab, referenceTab);
+          movedTabs.add(tab);
+          needToBeReindexedTabs.add(tab);
+        }
+        if (lastNextOfMovedTab)
+          needToBeReindexedTabs.add(lastNextOfMovedTab);
+        break;
+    }
+  }
+
+  //const currentIndices = Array.from(container.childNodes, tab => tab.apiTab.index);
+  //const reindexOperations = (new SequenceMatcher(currentIndices, apiTabIds.map((_id, index) => index))).operations();
+  //log('syncTabsOrder: reindexOperations ', reindexOperations);
+
+  const reindexTabs = Tabs.sort(Array.from(needToBeReindexedTabs));
+  const first = reindexTabs[0];
+  const last = reindexTabs[reindexTabs.length - 1];
+  const reindexedTabs = new Set();
+  const allTabs = container.childNodes;
+  if (reindexTabs.length > 0) {
+    log('syncTabsOrder: reindex between ', { first: first.apiTab.id, last: last.apiTab.id });
+    const lastCorrectIndexTab = Tabs.getPreviousTab(first);
+    for (let i = lastCorrectIndexTab ? lastCorrectIndexTab.apiTab.index + 1 : 0, maxi = allTabs.length; i < maxi; i++) {
+      const tab = allTabs[i];
+      tab.apiTab.index = i;
+      reindexedTabs.add(tab);
+      if (tab == last)
+        break;
+    }
+  }
+
+  if (movedTabs.size > 0 || reindexedTabs.size > 0)
+    log('Tab nodes rearranged and reindexed by syncTabsOrder:\n'+(!configs.debug ? '' :
+      Array.from(allTabs)
+        .map(tab => ' - '+tab.apiTab.index+': '+tab.id+(movedTabs.has(tab) ? '[MOVED]' : '')+(reindexedTabs.has(tab) ? '[REINDEXED]' : '')+' '+tab.apiTab.title)
+        .join('\n')));
+}
+
+
+
 Tabs.onBuilt.addListener((tab, info) => {
   const label = Tabs.getTabLabel(tab);
 
@@ -248,7 +447,7 @@ Tabs.onBuilt.addListener((tab, info) => {
   tab.appendChild(burster);
 
   const activeMarker = document.createElement('span');
-  activeMarker.classList.add(Constants.kACTIVE_MARKER);
+  activeMarker.classList.add(Constants.kHIGHLIGHTER);
   tab.appendChild(activeMarker);
 
   const identityMarker = document.createElement('span');
@@ -274,6 +473,23 @@ Tabs.onCreated.addListener((tab, _info) => {
   tab.classList.add(Constants.kTAB_STATE_ANIMATION_READY);
 });
 
+Tabs.onTabElementMoved.addListener((tab, info) => {
+  if (!info.broadcasted) {
+    // Tab element movement triggered by sidebar itself can break order of
+    // tabs synchronized from the background, so for safetyl we trigger
+    // synchronization.
+    reserveToSyncTabsOrder();
+  }
+});
+
+
+Tabs.onRestored.addListener(tab => {
+  Tree.fixupSubtreeCollapsedState(tab, {
+    justNow:  true,
+    inRemote: true
+  });
+});
+
 Tabs.onRemoving.addListener((_tab, _info) => {
   reserveToUpdateLoadingState();
 });
@@ -336,6 +552,8 @@ Tabs.onStateChanged.addListener(tab => {
 
 Tabs.onLabelUpdated.addListener(tab => {
   reserveToUpdateTooltip(tab);
+  if (!tab.titleUpdatedWhileCollapsed && Tabs.isCollapsed(tab))
+    tab.titleUpdatedWhileCollapsed = true;
 });
 
 Tabs.onFaviconUpdated.addListener((tab, url) => {
@@ -346,31 +564,43 @@ Tabs.onFaviconUpdated.addListener((tab, url) => {
   });
 });
 
-Tabs.onCollapsedStateChanged.addListener((tab, _info) => { reserveToUpdateLoadingState(tab); });
+Tabs.onCollapsedStateChanged.addListener((tab, info) => {
+  if (info.collapsed)
+    return;
+  reserveToUpdateLoadingState(tab);
+  if (tab.titleUpdatedWhileCollapsed) {
+    updateLabelOverflow(tab);
+    delete tab.titleUpdatedWhileCollapsed;
+  }
+});
 
-const mDelayedUpdateForHighlightChange = new WeakMap();
-
-Tabs.onUpdated.addListener(async (tab, info) => {
-  updateSoundButtonTooltip(tab);
+let mReservedUpdateActiveTab;
+Tabs.onUpdated.addListener((tab, info) => {
+  reserveToUpdateSoundButtonTooltip(tab);
+  reserveToUpdateTooltip(tab);
 
   if (!('highlighted' in info))
     return;
 
-  updateClosebox(tab);
+  reserveToUpdateCloseboxTooltip(tab);
 
-  const activeTab = Tabs.getCurrentTab(tab);
-  let timer = mDelayedUpdateForHighlightChange.get(activeTab);
-  if (activeTab)
-    clearTimeout(activeTab);
-  timer = setTimeout(() => {
-    mDelayedUpdateForHighlightChange.delete(activeTab);
-    updateSoundButtonTooltip(activeTab);
-    updateClosebox(activeTab);
-  }, 100);
-  mDelayedUpdateForHighlightChange.set(activeTab, timer);
+  for (const ancestor of Tabs.getAncestorTabs(tab)) {
+    updateDescendantsHighlighted(ancestor);
+  }
+
+  if (mReservedUpdateActiveTab)
+    clearTimeout(mReservedUpdateActiveTab);
+  mReservedUpdateActiveTab = setTimeout(() => {
+    mReservedUpdateActiveTab = null;
+    const activeTab = Tabs.getActiveTab();
+    if (activeTab) {
+      reserveToUpdateSoundButtonTooltip(activeTab);
+      reserveToUpdateCloseboxTooltip(activeTab);
+    }
+  }, 50);
 });
 
-Tabs.onParentTabUpdated.addListener(tab => { updateSoundButtonTooltip(tab); });
+Tabs.onParentTabUpdated.addListener(tab => { reserveToUpdateSoundButtonTooltip(tab); });
 
 Tabs.onDetached.addListener((tab, _info) => {
   if (!mInitialized ||
@@ -398,12 +628,13 @@ Tabs.onGroupTabDetected.addListener(tab => {
 Tree.onAttached.addListener((tab, info = {}) => {
   if (!mInitialized)
     return;
-  updateTwisty(info.parent);
-  updateClosebox(info.parent);
+  reserveToUpdateTwistyTooltip(info.parent);
+  reserveToUpdateCloseboxTooltip(info.parent);
   if (info.newlyAttached) {
     const ancestors = [info.parent].concat(Tabs.getAncestorTabs(info.parent));
     for (const ancestor of ancestors) {
       updateDescendantsCount(ancestor);
+      updateDescendantsHighlighted(ancestor);
     }
   }
   reserveToUpdateTooltip(info.parent);
@@ -415,18 +646,19 @@ Tree.onDetached.addListener((_aTab, detachInfo = {}) => {
   const parent = detachInfo.oldParentTab;
   if (!parent)
     return;
-  updateTwisty(parent);
-  updateClosebox(parent);
+  reserveToUpdateTwistyTooltip(parent);
+  reserveToUpdateCloseboxTooltip(parent);
   reserveToUpdateTooltip(parent);
   const ancestors = [parent].concat(Tabs.getAncestorTabs(parent));
   for (const ancestor of ancestors) {
     updateDescendantsCount(ancestor);
+    updateDescendantsHighlighted(ancestor);
   }
 });
 
 Tree.onSubtreeCollapsedStateChanging.addListener((tab, _info) => {
-  updateTwisty(tab);
-  updateClosebox(tab);
+  reserveToUpdateTwistyTooltip(tab);
+  reserveToUpdateCloseboxTooltip(tab);
   if (mInitialized)
     reserveToUpdateTooltip(tab);
 });
