@@ -24,6 +24,7 @@ import * as Tree from '/common/tree.js';
 import * as TSTAPI from '/common/tst-api.js';
 import * as ContextualIdentities from '/common/contextual-identities.js';
 import * as Commands from '/common/commands.js';
+import * as Bookmark from '/common/bookmark.js';
 import * as UserOperationBlocker from '/common/user-operation-blocker.js';
 import * as MetricsData from '/common/metrics-data.js';
 
@@ -31,6 +32,7 @@ import * as SidebarCache from './sidebar-cache.js';
 import * as SidebarTabs from './sidebar-tabs.js';
 import * as PinnedTabs from './pinned-tabs.js';
 import * as DragAndDrop from './drag-and-drop.js';
+import * as TabDragHandle from './tab-drag-handle.js';
 import * as RestoringTabCount from './restoring-tab-count.js';
 import * as Size from './size.js';
 import * as Color from './color.js';
@@ -118,7 +120,6 @@ export async function init() {
 
       restoredFromCache = await rebuildAll(cachedContents && cachedContents.tabbar);
       ApiTabsListener.startListen();
-      TabsUpdate.startListen();
 
       browser.runtime.connect({
         name: `${Constants.kCOMMAND_REQUEST_CONNECT_PREFIX}${mTargetWindow}`
@@ -155,6 +156,7 @@ export async function init() {
       onBuilt.dispatch();
 
       DragAndDrop.init();
+      TabDragHandle.init();
 
       MetricsData.add('onBuilt: start to listen events');
     }),
@@ -165,7 +167,6 @@ export async function init() {
     MetricsData.addAsync('initializing contextual identities', async () => {
       updateContextualIdentitiesStyle();
       updateContextualIdentitiesSelector();
-      updateContextualIdentitiesSelectorInContextMenu();
       ContextualIdentities.startObserve();
     }),
     MetricsData.addAsync('TabContextMenu.init', async () => {
@@ -274,12 +275,13 @@ function applyBrowserTheme(aTheme) {
   if (aTheme.colors.toolbar)
     toolbarColor = Color.mixCSSColors(themeBaseColor, aTheme.colors.toolbar);
   if (aTheme.colors.tab_line)
-    extraColors.push(`--browser-tab-active-marker: ${aTheme.colors.tab_line}`);
+    extraColors.push(`--browser-tab-highlighter: ${aTheme.colors.tab_line}`);
   if (aTheme.colors.tab_loading)
     extraColors.push(`--browser-loading-indicator: ${aTheme.colors.tab_loading}`);
   mBrowserThemeDefinition.textContent = `
     ${defaultColors}
     :root {
+      --browser-background:      ${aTheme.colors.accentcolor};
       --browser-bg-base:         ${themeBaseColor};
       --browser-bg-less-lighter: ${Color.mixCSSColors(themeBaseColor, 'rgba(255, 255, 255, 0.25)', bgAlpha)};
       --browser-bg-lighter:      ${toolbarColor};
@@ -311,7 +313,7 @@ function updateContextualIdentitiesStyle() {
 }
 
 function updateContextualIdentitiesSelector() {
-  const anchors = Array.slice(document.querySelectorAll(`.${Constants.kCONTEXTUAL_IDENTITY_SELECTOR}-marker`));
+  const anchors = document.querySelectorAll(`.${Constants.kCONTEXTUAL_IDENTITY_SELECTOR}-marker`);
   for (const anchor of anchors) {
     if (ContextualIdentities.getCount() == 0)
       anchor.setAttribute('disabled', true);
@@ -330,7 +332,6 @@ function updateContextualIdentitiesSelector() {
     item.dataset.value = identity.cookieStoreId;
     item.textContent = identity.name;
     item.dataset.icon = identity.iconUrl;
-    item.dataset.iconColor = identity.colorCode || 'var(--tab-text)';
     fragment.appendChild(item);
   });
   if (configs.inheritContextualIdentityToNewChildTab) {
@@ -346,35 +347,6 @@ function updateContextualIdentitiesSelector() {
   range.detach();
 }
 
-function updateContextualIdentitiesSelectorInContextMenu() {
-  const container = document.getElementById(Constants.kCONTEXTUAL_IDENTITY_SELECTOR_CONTEXT_MENU);
-
-  const range = document.createRange();
-  range.selectNodeContents(container);
-  range.setStartAfter(container.querySelector('.separator'));
-  range.deleteContents();
-
-  let hasIdentity = false;
-  const fragment = document.createDocumentFragment();
-  ContextualIdentities.forEach(identity => {
-    const item = document.createElement('li');
-    item.id = `context_reopenInContainer:${identity.cookieStoreId}`
-    item.dataset.value = identity.cookieStoreId;
-    item.textContent = identity.name.replace(/^([a-z0-9])/i, '&$1');
-    item.setAttribute('title', identity.name);
-    if (identity.iconUrl) {
-      item.dataset.icon = identity.iconUrl;
-      item.dataset.iconColor = identity.colorCode || 'var(--tab-text)';
-    }
-    fragment.appendChild(item);
-    hasIdentity = true;
-  });
-  range.insertNode(fragment);
-  range.detach();
-
-  container.parentNode.style.display = hasIdentity ? '' : 'none' ;
-}
-
 export async function rebuildAll(cache) {
   const apiTabs = await browser.tabs.query({ currentWindow: true });
   TabsContainer.clearAll();
@@ -435,10 +407,13 @@ async function waitUntilBackgroundIsReady() {
 }
 
 
-export async function confirmToCloseTabs(count, _aOptions = {}) {
+export async function confirmToCloseTabs(apiTabIds, _options = {}) {
+  apiTabIds = apiTabIds.filter(id => !configs.grantedRemovingTabIds.includes(id));
+  const count = apiTabIds.length;
   if (count <= 1 ||
       !configs.warnOnCloseTabs)
     return true;
+
   const confirm = new RichConfirm({
     message: browser.i18n.getMessage('warnOnCloseTabs_message', [count]),
     buttons: [
@@ -454,6 +429,8 @@ export async function confirmToCloseTabs(count, _aOptions = {}) {
       if (!result.checked)
         configs.warnOnCloseTabs = false;
       configs.lastConfirmedToCloseTabs = Date.now();
+      configs.grantedRemovingTabIds = Array.from(new Set((configs.grantedRemovingTabIds || []).concat(apiTabIds)));
+      log('confirmToCloseTabs: granted ', configs.grantedRemovingTabIds);
       return true;
     default:
       return false;
@@ -510,7 +487,7 @@ function updateTabbarLayout(params = {}) {
       // Tab at the end of the tab bar can be hidden completely or
       // partially (newly opened in small tab bar, or scrolled out when
       // the window is shrunken), so we need to scroll to it explicitely.
-      const current = Tabs.getCurrentTab();
+      const current = Tabs.getActiveTab();
       if (!Scroll.isTabInViewport(current)) {
         log('scroll to current tab on updateTabbarLayout');
         Scroll.scrollToTab(current);
@@ -586,8 +563,11 @@ Tabs.onCreated.addListener((_tab, _info) => {
   });
 });
 
-Tabs.onRemoving.addListener((tab, closeInfo) => {
-  const closeParentBehavior = Tree.getCloseParentBehaviorForTabWithSidebarOpenState(tab, closeInfo);
+Tabs.onRemoving.addListener((tab, removeInfo) => {
+  if (removeInfo.isWindowClosing)
+    return;
+
+  const closeParentBehavior = Tree.getCloseParentBehaviorForTabWithSidebarOpenState(tab, removeInfo);
   if (closeParentBehavior != Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN &&
       Tabs.isSubtreeCollapsed(tab))
     Tree.collapseExpandSubtree(tab, {
@@ -683,7 +663,6 @@ Tabs.onWindowRestoring.addListener(async windowId => {
 ContextualIdentities.onUpdated.addListener(() => {
   updateContextualIdentitiesStyle();
   updateContextualIdentitiesSelector();
-  updateContextualIdentitiesSelectorInContextMenu();
 });
 
 
@@ -893,6 +872,10 @@ function onMessage(message, _sender, _respond) {
       })();
     }; break;
 
+    case Constants.kCOMMAND_SYNC_TABS_ORDER:
+      SidebarTabs.reserveToSyncTabsOrder();
+      break;
+
     case Constants.kCOMMAND_MOVE_TABS_BEFORE:
       return (async () => {
         await Tabs.waitUntilTabsAreCreated(message.tabs.concat([message.nextTab]));
@@ -900,7 +883,12 @@ function onMessage(message, _sender, _respond) {
           message.tabs.map(Tabs.getTabById),
           Tabs.getTabById(message.nextTab),
           message
-        ).then(tabs => tabs.map(tab => tab.id));
+        ).then(tabs => {
+          // Asynchronously broadcasted movement can break the order of tabs,
+          // so we trigger synchronization for safety.
+          SidebarTabs.reserveToSyncTabsOrder();
+          return tabs.map(tab => tab.id);
+        });
       })();
 
     case Constants.kCOMMAND_MOVE_TABS_AFTER:
@@ -910,7 +898,12 @@ function onMessage(message, _sender, _respond) {
           message.tabs.map(Tabs.getTabById),
           Tabs.getTabById(message.previousTab),
           message
-        ).then(tabs => tabs.map(tab => tab.id));
+        ).then(tabs => {
+          // Asynchronously broadcasted movement can break the order of tabs,
+          // so we trigger synchronization for safety.
+          SidebarTabs.reserveToSyncTabsOrder();
+          return tabs.map(tab => tab.id);
+        });
       })();
 
     case Constants.kCOMMAND_REMOVE_TABS_INTERNALLY:
@@ -996,7 +989,7 @@ function onMessage(message, _sender, _respond) {
           if (modified.includes(Constants.kTAB_STATE_AUDIBLE) ||
             modified.includes(Constants.kTAB_STATE_SOUND_PLAYING) ||
             modified.includes(Constants.kTAB_STATE_MUTED)) {
-            SidebarTabs.updateSoundButtonTooltip(tab);
+            SidebarTabs.reserveToUpdateSoundButtonTooltip(tab);
             if (message.bubbles)
               TabsUpdate.updateParentTab(Tabs.getParentTab(tab));
           }
@@ -1005,8 +998,20 @@ function onMessage(message, _sender, _respond) {
     }; break;
 
     case Constants.kCOMMAND_CONFIRM_TO_CLOSE_TABS: {
+      log('kCOMMAND_CONFIRM_TO_CLOSE_TABS: ', { message, mTargetWindow });
       if (message.windowId == mTargetWindow)
-        return confirmToCloseTabs(message.count);
+        return confirmToCloseTabs(message.tabIds);
     }; break;
+
+
+    case Constants.kCOMMAND_BOOKMARK_TAB_WITH_DIALOG:
+      if (message.windowId != mTargetWindow)
+        return;
+      return Bookmark.bookmarkTab(Tabs.getTabById(message.tab), { showDialog: true });
+
+    case Constants.kCOMMAND_BOOKMARK_TABS_WITH_DIALOG:
+      if (message.windowId != mTargetWindow)
+        return;
+      return Bookmark.bookmarkTabs(message.tabs.map(Tabs.getTabById), { showDialog: true });
   }
 }
