@@ -70,6 +70,23 @@ export function getWindow() {
   return mTargetWindow;
 }
 
+export function sort(tabs) {
+  return tabs.sort(documentPositionComparator);
+}
+
+function documentPositionComparator(a, b) {
+  if (a === b || !a || !b)
+    return 0;
+
+  const position = a.compareDocumentPosition(b);
+  if (position & Node.DOCUMENT_POSITION_FOLLOWING)
+    return -1;
+  if (position & Node.DOCUMENT_POSITION_PRECEDING)
+    return 1;
+
+  return 0;
+}
+
 
 //===================================================================
 // Operate Tab ID
@@ -202,74 +219,153 @@ export const onWindowRestoring  = new EventListenerManager();
 export const onAttached         = new EventListenerManager();
 export const onDetached         = new EventListenerManager();
 
-async function waitUntilTabsAreOperated(idOrIds, slot) {
+function normalizeOperatingTabIds(idOrIds) {
   if (!Array.isArray(idOrIds))
     idOrIds = [idOrIds];
-  const operatingTabs = idOrIds
+  return idOrIds
     .map(id => parseInt(id))
     .filter(id => !!id)
-    .map(id => typeof id == 'string' ? parseInt(id.match(/^tab-\d+-(\d+)$/)[1]) : id)
-    .map(id => slot.get(id))
-    .filter(operating => !!operating);
-  if (operatingTabs.length)
-    return Promise.all(operatingTabs);
+    .map(id => typeof id == 'string' ? parseInt(id.match(/^tab-\d+-(\d+)$/)[1]) : id);
+}
+
+async function waitUntilTabsAreOperated(params = {}) {
+  const ids = params.ids && normalizeOperatingTabIds(params.ids);
+  let promises = [];
+  if (params.operatingTabsInWindow) {
+    if (ids) {
+      for (const id of ids) {
+        if (params.operatingTabsInWindow.has(id))
+          promises.push(params.operatingTabsInWindow.get(id));
+      }
+    }
+    else {
+      promises.splice(0, 0, ...params.operatingTabsInWindow.values());
+    }
+  }
+  else if (params.operatingTabs) {
+    for (const operatingTabsInWindow of params.operatingTabs.values()) {
+      if (ids) {
+        for (let i = ids.length - 1; i > -1; i--) {
+          const id = ids[i];
+          if (operatingTabsInWindow.has(id)) {
+            promises.push(operatingTabsInWindow.get(id));
+            ids.splice(i, 1);
+          }
+        }
+        if (ids.length == 0)
+          break;
+      }
+      else {
+        promises.splice(0, 0, ...operatingTabsInWindow.values());
+      }
+    }
+  }
+  else {
+    throw new Error('missing required parameter: operatingTabs or operatingTabsInWindow');
+  }
+  promises = promises.filter(operating => !!operating);
+  if (promises.length > 0)
+    return Promise.all(promises);
   return [];
 }
 
+export function hasOperatingTab(params = {}) {
+  if (!params.operatingTabs) {
+    throw new Error('missing required parameter: operatingTabs');
+  }
+  if (params.windowId) {
+    const operatingTabsInWindow = params.operatingTabs.get(params.windowId);
+    return operatingTabsInWindow ? operatingTabsInWindow.size > 0 : false;
+  }
+  for (const operatingTabsInWindow of params.operatingTabs.values()) {
+    if (operatingTabsInWindow.size > 0)
+      return true;
+  }
+  return false;
+}
+
 const mCreatingTabs = new Map();
 
 export function addCreatingTab(tab) {
   let onTabCreated;
+  const creatingTabs = mCreatingTabs.get(tab.apiTab.windowId) || new Map();
   if (configs.acceleratedTabCreation) {
-    mCreatingTabs.set(tab.apiTab.id, tab.uniqueId);
+    creatingTabs.set(tab.apiTab.id, tab.uniqueId);
     onTabCreated = () => {};
   }
   else {
-    mCreatingTabs.set(tab.apiTab.id, new Promise((resolve, _aReject) => {
+    creatingTabs.set(tab.apiTab.id, new Promise((resolve, _aReject) => {
       onTabCreated = (uniqueId) => { resolve(uniqueId); };
     }));
   }
+  mCreatingTabs.set(tab.apiTab.windowId, creatingTabs);
   tab.uniqueId.then(_aUniqueId => {
-    mCreatingTabs.delete(tab.apiTab.id);
+    creatingTabs.delete(tab.apiTab.id);
   });
   return onTabCreated;
 }
 
-export function hasCreatingTab() {
-  return mCreatingTabs.size > 0;
+export function hasCreatingTab(windowId = null) {
+  return hasOperatingTab({ operatingTabs: mCreatingTabs, windowId });
 }
 
-export async function waitUntilAllTabsAreCreated() {
-  return waitUntilTabsAreCreated(Array.from(mCreatingTabs.keys()));
+export async function waitUntilAllTabsAreCreated(windowId = null) {
+  const params = {};
+  if (windowId) {
+    params.operatingTabsInWindow = mCreatingTabs.get(windowId);
+    if (!params.operatingTabsInWindow)
+      return;
+  }
+  else {
+    params.operatingTabs = mCreatingTabs;
+  }
+  return waitUntilTabsAreOperated(params)
+    .then(aUniqueIds => aUniqueIds.map(uniqueId => getTabByUniqueId(uniqueId.id)));
 }
 
 export async function waitUntilTabsAreCreated(idOrIds) {
-  return waitUntilTabsAreOperated(idOrIds, mCreatingTabs)
+  return waitUntilTabsAreOperated({ ids: idOrIds, operatingTabs: mCreatingTabs })
     .then(aUniqueIds => aUniqueIds.map(uniqueId => getTabByUniqueId(uniqueId.id)));
 }
 
 const mMovingTabs = new Map();
 
-export function hasMovingTab() {
-  return mMovingTabs.size > 0;
+export function hasMovingTab(windowId = null) {
+  return hasOperatingTab({ operatingTabs: mMovingTabs, windowId });
 }
 
-export function addMovingTabId(tabId) {
+export function addMovingTabId(tabId, windowId) {
   let onTabMoved;
   const promisedMoved = new Promise((resolve, _aReject) => {
     onTabMoved = resolve;
   });
-  mMovingTabs.set(tabId, promisedMoved);
+  const movingTabs = mMovingTabs.get(windowId) || new Map();
+  movingTabs.set(tabId, promisedMoved);
+  mMovingTabs.set(windowId, movingTabs);
   promisedMoved.then(() => {
-    mMovingTabs.delete(tabId);
+    movingTabs.delete(tabId);
   });
   return onTabMoved;
 }
 
-export async function waitUntilAllTabsAreMoved() {
-  return waitUntilTabsAreOperated(Array.from(mMovingTabs.keys()), mMovingTabs);
+export async function waitUntilAllTabsAreMoved(windowId = null) {
+  const params = {};
+  if (windowId) {
+    params.operatingTabsInWindow = mMovingTabs.get(windowId);
+    if (!params.operatingTabsInWindow)
+      return;
+  }
+  else {
+    params.operatingTabs = mMovingTabs;
+  }
+  return waitUntilTabsAreOperated(params)
 }
 
+browser.windows.onRemoved.addListener(windowId => {
+  mCreatingTabs.delete(windowId);
+  mMovingTabs.delete(windowId);
+});
+
 
 //===================================================================
 // Create Tabs
@@ -426,12 +522,12 @@ export function getTabLabelContent(tab) {
 
 // Note that this function can return null if it is the first tab of
 // a new window opened by the "move tab to new window" command.
-export function getCurrentTab(hint) {
+export function getActiveTab(hint) {
   const container = getTabsContainer(hint);
   return container && container.querySelector(`.${Constants.kTAB_STATE_ACTIVE}`);
 }
-export function getCurrentTabs() {
-  return Array.slice(document.querySelectorAll(`.${Constants.kTAB_STATE_ACTIVE}`));
+export function getActiveTabs() {
+  return Array.from(document.querySelectorAll(`.${Constants.kTAB_STATE_ACTIVE}`));
 }
 
 export function getNextTab(tab) {
@@ -702,35 +798,35 @@ export function getAllTabs(hint) {
   const container = getTabsContainer(hint);
   if (!container)
     return [];
-  return Array.slice(container.querySelectorAll(kSELECTOR_LIVE_TAB));
+  return Array.from(container.querySelectorAll(kSELECTOR_LIVE_TAB));
 }
 
 export function getTabs(hint) { // only visible, including collapsed and pinned
   const container = getTabsContainer(hint);
   if (!container)
     return [];
-  return Array.slice(container.querySelectorAll(kSELECTOR_CONTROLLABLE_TAB));
+  return Array.from(container.querySelectorAll(kSELECTOR_CONTROLLABLE_TAB));
 }
 
 export function getNormalTabs(hint) { // only visible, including collapsed, not pinned
   const container = getTabsContainer(hint);
   if (!container)
     return [];
-  return Array.slice(container.querySelectorAll(kSELECTOR_NORMAL_TAB));
+  return Array.from(container.querySelectorAll(kSELECTOR_NORMAL_TAB));
 }
 
-function getVisibleTabs(hint) { // visible, not-collapsed, not-hidden
+export function getVisibleTabs(hint) { // visible, not-collapsed, not-hidden
   const container = getTabsContainer(hint);
   if (!container)
     return [];
-  return Array.slice(container.querySelectorAll(kSELECTOR_VISIBLE_TAB));
+  return Array.from(container.querySelectorAll(kSELECTOR_VISIBLE_TAB));
 }
 
 export function getPinnedTabs(hint) { // visible, pinned
   const container = getTabsContainer(hint);
   if (!container)
     return [];
-  return Array.slice(container.querySelectorAll(kSELECTOR_PINNED_TAB));
+  return Array.from(container.querySelectorAll(kSELECTOR_PINNED_TAB));
 }
 
 
@@ -738,32 +834,32 @@ export function getUnpinnedTabs(hint) { // visible, not pinned
   const container = getTabsContainer(hint);
   if (!container)
     return [];
-  return Array.slice(container.querySelectorAll(`${kSELECTOR_LIVE_TAB}:not(.${Constants.kTAB_STATE_PINNED})`));
+  return Array.from(container.querySelectorAll(`${kSELECTOR_LIVE_TAB}:not(.${Constants.kTAB_STATE_PINNED})`));
 }
 
 /*
 function getAllRootTabs(hint) {
   const container = getTabsContainer(hint);
-  return Array.slice(container.querySelectorAll(`${kSELECTOR_LIVE_TAB}:not([${Constants.kPARENT}])`));
+  return Array.from(container.querySelectorAll(`${kSELECTOR_LIVE_TAB}:not([${Constants.kPARENT}])`));
 }
 */
 
 export function getRootTabs(hint) {
   const container = getTabsContainer(hint);
-  return Array.slice(container.querySelectorAll(`${kSELECTOR_CONTROLLABLE_TAB}:not([${Constants.kPARENT}])`));
+  return Array.from(container.querySelectorAll(`${kSELECTOR_CONTROLLABLE_TAB}:not([${Constants.kPARENT}])`));
 }
 
 /*
 function getVisibleRootTabs(hint) {
   const container = getTabsContainer(hint);
-  return Array.slice(container.querySelectorAll(`${kSELECTOR_VISIBLE_TAB}:not([${Constants.kPARENT}])`));
+  return Array.from(container.querySelectorAll(`${kSELECTOR_VISIBLE_TAB}:not([${Constants.kPARENT}])`));
 }
 
 function getVisibleLoadingTabs(hint) {
   const container = getTabsContainer(hint);
   if (!container)
     return [];
-  return Array.slice(container.querySelectorAll(`${kSELECTOR_VISIBLE_TAB}.loading`));
+  return Array.from(container.querySelectorAll(`${kSELECTOR_VISIBLE_TAB}.loading`));
 }
 */
 
@@ -779,7 +875,7 @@ export function collectRootTabs(tabs) {
 /*
 function getIndentedTabs(hint) {
   const container = getTabsContainer(hint);
-  return Array.slice(container.querySelectorAll(`${kSELECTOR_CONTROLLABLE_TAB}[${Constants.kPARENT}]`));
+  return Array.from(container.querySelectorAll(`${kSELECTOR_CONTROLLABLE_TAB}[${Constants.kPARENT}]`));
 }
 
 function getVisibleIndentedTabs(hint) {
@@ -790,19 +886,28 @@ function getVisibleIndentedTabs(hint) {
 
 export function getDraggingTabs(hint) {
   const container = getTabsContainer(hint);
-  return Array.slice(container.querySelectorAll(`${kSELECTOR_LIVE_TAB}.${Constants.kTAB_STATE_DRAGGING}`));
+  return Array.from(container.querySelectorAll(`${kSELECTOR_LIVE_TAB}.${Constants.kTAB_STATE_DRAGGING}`));
 }
 
 export function getDuplicatingTabs(hint) {
   const container = getTabsContainer(hint);
-  return Array.slice(container.querySelectorAll(`${kSELECTOR_LIVE_TAB}.${Constants.kTAB_STATE_DUPLICATING}`));
+  return Array.from(container.querySelectorAll(`${kSELECTOR_LIVE_TAB}.${Constants.kTAB_STATE_DUPLICATING}`));
+}
+
+export function getHighlightedTabs(hint) {
+  const container = getTabsContainer(hint);
+  if (!container)
+    return [];
+  return Array.from(container.querySelectorAll(`
+    ${kSELECTOR_LIVE_TAB}.${Constants.kTAB_STATE_HIGHLIGHTED}
+  `));
 }
 
 export function getSelectedTabs(hint) {
   const container = getTabsContainer(hint);
   if (!container)
     return [];
-  return Array.slice(container.querySelectorAll(`
+  return Array.from(container.querySelectorAll(`
     ${kSELECTOR_LIVE_TAB}.${Constants.kTAB_STATE_SELECTED},
     .${Constants.kTABBAR_STATE_MULTIPLE_HIGHLIGHTED} ${kSELECTOR_LIVE_TAB}.${Constants.kTAB_STATE_HIGHLIGHTED}
   `));
@@ -892,7 +997,7 @@ export async function doAndGetNewTabs(asyncTask, hint) {
   return addedTabs;
 }
 
-export function getNextFocusedTab(tab, options = {}) { // if the current tab is closed...
+export function getNextActiveTab(tab, options = {}) { // if the current tab is closed...
   const ignoredTabs = (options.ignoredTabs || []).slice(0);
   let foundTab = tab;
   do {
@@ -935,46 +1040,46 @@ export function getOpenerFromGroupTab(groupTab) {
 
 export function isActive(tab) {
   return ensureLivingTab(tab) &&
-           tab.classList.contains(Constants.kTAB_STATE_ACTIVE);
+           !!(tab.apiTab && tab.apiTab.active);
 }
 
 export function isPinned(tab) {
   return ensureLivingTab(tab) &&
-           tab.classList.contains(Constants.kTAB_STATE_PINNED);
+           !!(tab.apiTab && tab.apiTab.pinned);
 }
 
 export function isAudible(tab) {
   return ensureLivingTab(tab) &&
-           tab.classList.contains(Constants.kTAB_STATE_AUDIBLE);
+           !!(tab.apiTab && tab.apiTab.audible);
 }
 
 export function isSoundPlaying(tab) {
   return ensureLivingTab(tab) &&
-           tab.classList.contains(Constants.kTAB_STATE_SOUND_PLAYING);
+           !!(tab.apiTab && tab.apiTab.audible && !tab.apiTab.mutedInfo.muted);
 }
 
 export function maybeSoundPlaying(tab) {
   return ensureLivingTab(tab) &&
-         (tab.classList.contains(Constants.kTAB_STATE_SOUND_PLAYING) ||
+         (isSoundPlaying(tab) ||
           (tab.classList.contains(Constants.kTAB_STATE_HAS_SOUND_PLAYING_MEMBER) &&
            tab.hasAttribute(Constants.kCHILDREN)));
 }
 
 export function isMuted(tab) {
   return ensureLivingTab(tab) &&
-           tab.classList.contains(Constants.kTAB_STATE_MUTED);
+           !!(tab.apiTab && tab.apiTab.mutedInfo && tab.apiTab.mutedInfo.muted);
 }
 
 export function maybeMuted(tab) {
   return ensureLivingTab(tab) &&
-         (tab.classList.contains(Constants.kTAB_STATE_MUTED) ||
+         (isMuted(tab) ||
           (tab.classList.contains(Constants.kTAB_STATE_HAS_MUTED_MEMBER) &&
            tab.hasAttribute(Constants.kCHILDREN)));
 }
 
 export function isHidden(tab) {
   return ensureLivingTab(tab) &&
-           tab.classList.contains(Constants.kTAB_STATE_HIDDEN);
+           !!(tab.apiTab && tab.apiTab.hidden);
 }
 
 export function isCollapsed(tab) {
@@ -984,12 +1089,12 @@ export function isCollapsed(tab) {
 
 export function isDiscarded(tab) {
   return ensureLivingTab(tab) &&
-           tab.classList.contains(Constants.kTAB_STATE_DISCARDED);
+           !!(tab.apiTab && tab.apiTab.discarded);
 }
 
 export function isPrivateBrowsing(tab) {
   return ensureLivingTab(tab) &&
-           tab.classList.contains(Constants.kTAB_STATE_PRIVATE_BROWSING);
+           !!(tab.apiTab && tab.apiTab.incognito);
 }
 
 export function isOpening(tab) {
@@ -1047,18 +1152,22 @@ export function isTemporaryGroupTab(tab) {
 export function isSelected(tab) {
   return ensureLivingTab(tab) &&
            (tab.classList.contains(Constants.kTAB_STATE_SELECTED) ||
-            tab.matches(`.${Constants.kTABBAR_STATE_MULTIPLE_HIGHLIGHTED} .${Constants.kTAB_STATE_HIGHLIGHTED}`));
+            (isMultihighlighted(tab) && !!(tab.apiTab && tab.apiTab.highlighted)));
 }
 
 export function isHighlighted(tab) {
   return ensureLivingTab(tab) &&
-           tab.classList.contains(Constants.kTAB_STATE_HIGHLIGHTED);
+           !!(tab.apiTab && tab.apiTab.highlighted);
 }
 
 export function isMultiselected(tab) {
   return isSelected(tab) &&
-           (tab.parentNode.matches(`.${Constants.kTABBAR_STATE_MULTIPLE_HIGHLIGHTED}`) ||
-            tab.parentNode.querySelector(`${kSELECTOR_LIVE_TAB}.${Constants.kTAB_STATE_SELECTED} ~ ${kSELECTOR_LIVE_TAB}.${Constants.kTAB_STATE_SELECTED}`));
+           (isMultihighlighted(tab) ||
+            !!tab.parentNode.querySelector(`${kSELECTOR_LIVE_TAB}.${Constants.kTAB_STATE_SELECTED} ~ ${kSELECTOR_LIVE_TAB}.${Constants.kTAB_STATE_SELECTED}`));
+}
+
+export function isMultihighlighted(tab) {
+  return tab.parentNode.matches(`.${Constants.kTABBAR_STATE_MULTIPLE_HIGHLIGHTED}`);
 }
 
 export function isLocked(_aTab) {
@@ -1096,7 +1205,7 @@ export function isAllTabsPlacedBefore(tabs, nextTab) {
   if (!nextTab && !getNextTab(tabs[tabs.length - 1]))
     return true;
 
-  tabs = Array.slice(tabs);
+  tabs = Array.from(tabs);
   let previousTab = tabs.shift();
   for (const tab of tabs) {
     if (tab.previousSibling != previousTab)
@@ -1105,7 +1214,7 @@ export function isAllTabsPlacedBefore(tabs, nextTab) {
   }
   return !nextTab ||
          !previousTab ||
-         previousTab.nextSibling == nextTab;
+         getNextTab(previousTab) == nextTab;
 }
 
 export function isAllTabsPlacedAfter(tabs, previousTab) {
@@ -1114,10 +1223,10 @@ export function isAllTabsPlacedAfter(tabs, previousTab) {
   if (!previousTab && !getPreviousTab(tabs[0]))
     return true;
 
-  tabs = Array.slice(tabs).reverse();
+  tabs = Array.from(tabs).reverse();
   let nextTab = tabs.shift();
   for (const tab of tabs) {
-    if (tab.nextSibling != nextTab)
+    if (getNextTab(tab) != nextTab)
       return false;
     nextTab = tab;
   }
@@ -1260,7 +1369,7 @@ function snapshotTree(targetTab, tabs) {
     const previous = getPreviousNormalTab(tab);
     item.previous = previous && previous.id;
   }
-  const activeTab = getCurrentTab(targetTab);
+  const activeTab = getActiveTab(targetTab);
   return {
     target:   snapshotById[targetTab.id],
     active:   activeTab && snapshotById[activeTab.id],
