@@ -75,10 +75,13 @@ export const onLevelChanged = new EventListenerManager();
 export const onSubtreeCollapsedStateChanging = new EventListenerManager();
 
 
+// return moved (or not)
 export async function attachTabTo(child, parent, options = {}) {
+  parent = Tabs.ensureLivingTab(parent);
+  child = Tabs.ensureLivingTab(child);
   if (!parent || !child) {
     log('missing information: ', dumpTab(parent), dumpTab(child));
-    return;
+    return false;
   }
 
   log('attachTabTo: ', {
@@ -101,16 +104,16 @@ export async function attachTabTo(child, parent, options = {}) {
 
   if (Tabs.isPinned(parent) || Tabs.isPinned(child)) {
     log('=> pinned tabs cannot be attached');
-    return;
+    return false;
   }
   if (parent.apiTab.windowId != child.apiTab.windowId) {
     log('=> could not attach tab to a parent in different window');
-    return;
+    return false;
   }
   const ancestors = [parent].concat(Tabs.getAncestorTabs(child));
   if (ancestors.includes(child)) {
     log('=> canceled for recursive request');
-    return;
+    return false;
   }
 
   if (options.dontMove) {
@@ -125,18 +128,27 @@ export async function attachTabTo(child, parent, options = {}) {
     options.insertAfter  = refTabs.insertAfter;
   }
   options.insertAfter = options.insertAfter || parent;
-  log('reference tabs: ', {
+  log(`reference tabs for ${child.id}: `, {
     next: dumpTab(options.insertBefore),
     prev: dumpTab(options.insertAfter)
   });
 
-  await Tabs.waitUntilAllTabsAreCreated();
+  await Tabs.waitUntilAllTabsAreCreated(child.apiTab.windowId);
+
+  parent = Tabs.ensureLivingTab(parent);
+  child = Tabs.ensureLivingTab(child);
+  if (!parent || !child) {
+    log('attachTabTo: parent or child is closed before attaching.');
+    return false;
+  }
+
   const newIndex = Tabs.calculateNewTabIndex({
     insertBefore: options.insertBefore,
     insertAfter:  options.insertAfter,
     ignoreTabs:   [child]
   });
-  log('newIndex: ', newIndex);
+  const moved = newIndex != child.apiTab.index;
+  log(`newIndex for ${child.id}: `, newIndex);
 
   const newlyAttached = (
     !parent.childTabs.includes(child) ||
@@ -172,8 +184,10 @@ export async function attachTabTo(child, parent, options = {}) {
       broadcast: false
     }));
 
+    log('attachTabTo: setting child information: ', dumpTab(parent));
     parent.setAttribute(Constants.kCHILDREN, `|${childIds.join('|')}|`);
 
+    log('attachTabTo: setting parent information: ', dumpTab(child));
     child.setAttribute(Constants.kPARENT, parent.id);
     child.parentTab = parent;
     child.ancestorTabs = Tabs.getAncestorTabs(child, { force: true });
@@ -214,6 +228,8 @@ export async function attachTabTo(child, parent, options = {}) {
       stack:            new Error().stack
     });
   }
+
+  return moved;
 }
 
 export function getReferenceTabsForNewChild(child, parent, options = {}) {
@@ -291,27 +307,29 @@ export function getReferenceTabsForNewChild(child, parent, options = {}) {
 export function detachTab(child, options = {}) {
   log('detachTab: ', dumpTab(child), options,
       { stack: `${new Error().stack}\n${options.stack || ''}` });
-  const parent = Tabs.getParentTab(child);
+  // the "parent" option is used for removing child.
+  const parent = Tabs.ensureLivingTab(options.parent) || Tabs.getParentTab(child);
 
   if (!parent)
-    log('parent is already removed, or orphan tab');
+    log(` => parent(${child.getAttribute(Constants.kPARENT)}) is already removed, or orphan tab`);
 
   if (parent) {
     parent.childTabs = parent.childTabs.filter(tab => tab != child);
     const childIds = parent.childTabs.map(tab => tab.id);
     if (childIds.length == 0) {
       parent.removeAttribute(Constants.kCHILDREN);
-      log('no more child');
+      log(' => no more child');
     }
     else {
       parent.setAttribute(Constants.kCHILDREN, `|${childIds.join('|')}|`);
-      log('rest children: ', childIds);
+      log(' => rest children: ', childIds);
     }
     TabsUpdate.updateParentTab(parent);
   }
   child.removeAttribute(Constants.kPARENT);
   child.parentTab = null;
   child.ancestorTabs = [];
+  log('detachTab: parent information cleared: ', dumpTab(child));
 
   updateTabsIndent(child);
 
@@ -333,7 +351,7 @@ export function detachTab(child, options = {}) {
 export async function detachTabsFromTree(tabs, options = {}) {
   if (!Array.isArray(tabs))
     tabs = [tabs];
-  tabs = Array.slice(tabs).reverse();
+  tabs = Array.from(tabs).reverse();
   const promisedAttach = [];
   for (const tab of tabs) {
     const children = Tabs.getChildTabs(tab);
@@ -354,9 +372,12 @@ export async function detachTabsFromTree(tabs, options = {}) {
 }
 
 export function detachAllChildren(tab, options = {}) {
-  const children = Tabs.getChildTabs(tab);
+  log('detachAllChildren: ', dumpTab(tab));
+  // the "children" option is used for removing tab.
+  const children = options.children ? options.children.map(Tabs.ensureLivingTab) : Tabs.getChildTabs(tab);
   if (!children.length)
     return;
+  log(' => children to be detached: ', children.map(dumpTab));
 
   if (!('behavior' in options))
     options.behavior = Constants.kCLOSE_PARENT_BEHAVIOR_SIMPLY_DETACH_ALL_CHILDREN;
@@ -365,7 +386,8 @@ export function detachAllChildren(tab, options = {}) {
 
   options.dontUpdateInsertionPositionInfo = true;
 
-  const parent = Tabs.getParentTab(tab);
+  // the "parent" option is used for removing tab.
+  const parent = Tabs.ensureLivingTab(options.parent) || Tabs.getParentTab(tab);
   if (Tabs.isGroupTab(tab) &&
       Tabs.getTabs(tab).filter(tab => tab.removing).length == children.length) {
     options.behavior = Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN;
@@ -427,37 +449,45 @@ export function detachAllChildren(tab, options = {}) {
   }
 }
 
+// returns moved (or not)
 export async function behaveAutoAttachedTab(tab, options = {}) {
-  const baseTab = options.baseTab || Tabs.getCurrentTab(Tabs.getWindow() || tab);
+  const baseTab = options.baseTab || Tabs.getActiveTab(Tabs.getWindow() || tab);
   log('behaveAutoAttachedTab ', dumpTab(tab), dumpTab(baseTab), options);
+  if (Tabs.isPinned(baseTab)) {
+    if (!Tabs.isPinned(tab))
+      return false;
+    options.behavior = Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING;
+    log(' => override behavior for pinned tabs');
+  }
   switch (options.behavior) {
     default:
-      break;
+      return false;
 
     case Constants.kNEWTAB_OPEN_AS_ORPHAN:
+      log(' => kNEWTAB_OPEN_AS_ORPHAN');
       detachTab(tab, {
         inRemote:  options.inRemote,
         broadcast: options.broadcast
       });
       if (Tabs.getNextTab(tab))
-        await TabsMove.moveTabAfter(tab, Tabs.getLastTab(), {
+        return TabsMove.moveTabAfter(tab, Tabs.getLastTab(), {
           delayedMove: true,
           inRemote: options.inRemote
         });
-      break;
+      return false;
 
     case Constants.kNEWTAB_OPEN_AS_CHILD:
-      await attachTabTo(tab, baseTab, {
+      log(' => kNEWTAB_OPEN_AS_CHILD');
+      return attachTabTo(tab, baseTab, {
         dontMove:    options.dontMove || configs.insertNewChildAt == Constants.kINSERT_NO_CONTROL,
         forceExpand: true,
         delayedMove: true,
         inRemote:    options.inRemote,
         broadcast:   options.broadcast
       });
-      return true;
-      break;
 
     case Constants.kNEWTAB_OPEN_AS_SIBLING: {
+      log(' => kNEWTAB_OPEN_AS_SIBLING');
       const parent = Tabs.getParentTab(baseTab);
       if (parent) {
         await attachTabTo(tab, parent, {
@@ -465,52 +495,90 @@ export async function behaveAutoAttachedTab(tab, options = {}) {
           inRemote:  options.inRemote,
           broadcast: options.broadcast
         });
+        return true;
       }
       else {
         detachTab(tab, {
           inRemote:  options.inRemote,
           broadcast: options.broadcast
         });
-        await TabsMove.moveTabAfter(tab, Tabs.getLastTab(), {
+        return TabsMove.moveTabAfter(tab, Tabs.getLastTab(), {
           delayedMove: true,
           inRemote: options.inRemote
         });
       }
-      return true;
-    }; break;
+    };
 
     case Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING: {
+      log(' => kNEWTAB_OPEN_AS_NEXT_SIBLING');
       let nextSibling = Tabs.getNextSiblingTab(baseTab);
       if (nextSibling == tab)
         nextSibling = null;
       const parent = Tabs.getParentTab(baseTab);
-      if (parent)
-        await attachTabTo(tab, parent, {
+      if (parent) {
+        return attachTabTo(tab, parent, {
           insertBefore: nextSibling,
-          insertAfter:  Tabs.getLastDescendantTab(baseTab),
+          insertAfter:  Tabs.getLastDescendantTab(baseTab) || baseTab,
           delayedMove:  true,
           inRemote:     options.inRemote,
           broadcast:    options.broadcast
         });
+      }
       else {
         detachTab(tab, {
           inRemote:  options.inRemote,
           broadcast: options.broadcast
         });
         if (nextSibling)
-          await TabsMove.moveTabBefore(tab, nextSibling, {
+          return TabsMove.moveTabBefore(tab, nextSibling, {
             delayedMove: true,
             inRemote:  options.inRemote,
             broadcast: options.broadcast
           });
         else
-          await TabsMove.moveTabAfter(tab, Tabs.getLastDescendantTab(baseTab), {
+          return TabsMove.moveTabAfter(tab, Tabs.getLastDescendantTab(baseTab), {
             delayedMove: true,
             inRemote:  options.inRemote,
             broadcast: options.broadcast
           });
       }
-    }; break;
+    };
+  }
+}
+
+export async function behaveAutoAttachedTabs(tabs, options = {}) {
+  switch (options.behavior) {
+    default:
+      return false;
+
+    case Constants.kNEWTAB_OPEN_AS_ORPHAN:
+      if (options.baseTabs && !options.baseTab)
+        options.baseTab = options.baseTabs[options.baseTabs.length-1];
+      for (const tab of tabs) {
+        await behaveAutoAttachedTab(tab, options);
+      }
+      return false;
+
+    case Constants.kNEWTAB_OPEN_AS_CHILD: {
+      if (options.baseTabs && !options.baseTab)
+        options.baseTab = options.baseTabs[0];
+      let moved = false;
+      for (const tab of tabs) {
+        moved = (await behaveAutoAttachedTab(tab, options)) || moved;
+      }
+      return moved;
+    };
+
+    case Constants.kNEWTAB_OPEN_AS_SIBLING:
+    case Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING: {
+      if (options.baseTabs && !options.baseTab)
+        options.baseTab = options.baseTabs[options.baseTabs.length-1];
+      let moved = false;
+      for (const tab of tabs.reverse()) {
+        moved = (await behaveAutoAttachedTab(tab, options)) || moved;
+      }
+      return moved;
+    };
   }
 }
 
@@ -641,7 +709,7 @@ export function collapseExpandTabAndSubtree(tab, params = {}) {
   if (params.collapsed && Tabs.isActive(tab)) {
     const newSelection = Tabs.getVisibleAncestorOrSelf(tab);
     logCollapseExpand('current tab is going to be collapsed, switch to ', dumpTab(newSelection));
-    TabsInternalOperation.selectTab(newSelection, { silently: true });
+    TabsInternalOperation.activateTab(newSelection, { silently: true });
   }
 
   if (!Tabs.isSubtreeCollapsed(tab)) {
@@ -710,11 +778,11 @@ export function collapseExpandTreesIntelligentlyFor(tab, options = {}) {
 
   logCollapseExpand('collapseExpandTreesIntelligentlyFor');
   const container = Tabs.getTabsContainer(tab);
-  if (parseInt(container.dataset.doingIntelligentlyCollapseExpandCount) > 0) {
+  if (container.doingIntelligentlyCollapseExpandCount > 0) {
     logCollapseExpand('=> done by others');
     return;
   }
-  TabsContainer.incrementCounter(container, 'doingIntelligentlyCollapseExpandCount');
+  container.doingIntelligentlyCollapseExpandCount++;
 
   const expandedAncestors = `<${[tab].concat(Tabs.getAncestorTabs(tab))
     .map(ancestor => ancestor.id)
@@ -758,75 +826,107 @@ export function collapseExpandTreesIntelligentlyFor(tab, options = {}) {
   collapseExpandSubtree(tab, Object.assign({}, options, {
     collapsed: false
   }));
-  TabsContainer.decrementCounter(container, 'doingIntelligentlyCollapseExpandCount');
+  container.doingIntelligentlyCollapseExpandCount--;
+}
+
+export async function fixupSubtreeCollapsedState(tab, options = {}) {
+  let fixed = false;
+  if (!Tabs.hasChildTabs(tab))
+    return fixed;
+  const firstChild = Tabs.getFirstChildTab(tab);
+  const childrenCollapsed = Tabs.isCollapsed(firstChild);
+  const collapsedStateMismatched = Tabs.isSubtreeCollapsed(tab) != childrenCollapsed;
+  const nextIsFirstChild = Tabs.getNextTab(tab) == firstChild;
+  log('fixupSubtreeCollapsedState ', {
+    tab: tab.id,
+    childrenCollapsed,
+    collapsedStateMismatched,
+    nextIsFirstChild
+  });
+  if (collapsedStateMismatched) {
+    log(' => set collapsed state');
+    await collapseExpandSubtree(tab, Object.assign({}, options, {
+      collapsed: childrenCollapsed
+    }));
+    fixed = true;
+  }
+  if (!nextIsFirstChild) {
+    log(' => move child tabs');
+    await followDescendantsToMovedRoot(tab, options);
+    fixed = true;
+  }
+  return fixed;
 }
 
 
 // operate tabs based on tree information
 
 /*
+These functions are not used on Firefox 65 and later. We should remove
+them after Firefox 64 and older versions are completely outdated.
+
  * By https://bugzilla.mozilla.org/show_bug.cgi?id=1366290 when the
    current tab is closed, Firefox notifies tabs.onTabRemoved at first
    and tabs.onActivated at later.
- * Basically the next (right) tab will be focused when the current tab
+ * Basically the next (right) tab will be active when the current tab
    is closed, except the closed tab was the last tab.
    * If the closed current tab was the last tab, then the previous tab
-     is focused.
- * However, if the tab has "owner", it will be focused instead of the
+     is active.
+ * However, if the tab has "owner", it will be active instead of the
    right tab if `browser.tabs.selectOwnerOnClose` == `true`.
    * The owner tab must be one of preceding tabs, because Firefox never
      open tab leftside (by default).
-     So, if the next (right) tab is focused, it definitely caused by
+     So, if the next (right) tab is active, it definitely caused by
      the closing of the current tab - except "switch to tab" command
      from the location bar.
      https://bugzilla.mozilla.org/show_bug.cgi?id=1405262
      https://github.com/piroor/treestyletab/issues/1409
 
 So, if I ignore the bug 1405262 / issue #1409 case, "the next (right)
-tab is focused after the current (active) tab is closed" means that the
+tab is active after the current (active) tab is closed" means that the
 focus move is unintentional and TST can override it.
 */
-export function tryMoveFocusFromClosingCurrentTab(tab, options = {}) {
-  if (!configs.moveFocusInTreeForClosedCurrentTab)
+export function tryMoveFocusFromClosingActiveTab(tab, options = {}) {
+  if (configs.successorTabControlLevel != Constants.kSUCCESSOR_TAB_CONTROL_IN_TREE)
     return;
-  log('tryMoveFocusFromClosingCurrentTab', dumpTab(tab), options);
+  log('tryMoveFocusFromClosingActiveTab', dumpTab(tab), options);
   if (!options.wasActive && !Tabs.isActive(tab)) {
     log(' => not active tab');
     return;
   }
-  tab.parentNode.focusRedirectedForClosingCurrentTab = tryMoveFocusFromClosingCurrentTabOnFocusRedirected(tab, options);
+  tab.parentNode.focusRedirectedForClosingActiveTab = tryMoveFocusFromClosingActiveTabOnFocusRedirected(tab, options);
 }
-async function tryMoveFocusFromClosingCurrentTabOnFocusRedirected(tab, options = {}) {
-  if (!configs.moveFocusInTreeForClosedCurrentTab)
+async function tryMoveFocusFromClosingActiveTabOnFocusRedirected(tab, options = {}) {
+  if (configs.successorTabControlLevel != Constants.kSUCCESSOR_TAB_CONTROL_IN_TREE)
     return false;
-  log('tryMoveFocusFromClosingCurrentTabOnFocusRedirected ', dumpTab(tab), options);
+  log('tryMoveFocusFromClosingActiveTabOnFocusRedirected ', dumpTab(tab), options);
 
   // The tab can be closed while we waiting.
   // Thus we need to get tabs related to tab at first.
-  const params      = getTryMoveFocusFromClosingCurrentTabNowParams(tab, options.params);
+  const params      = getTryMoveFocusFromClosingActiveTabNowParams(tab, options.params);
   const nextTab     = Tabs.getNextTab(tab);
   const previousTab = Tabs.getPreviousTab(tab);
 
   await tab.closedWhileActive;
-  log('tryMoveFocusFromClosingCurrentTabOnFocusRedirected: tabs.onActivated is fired');
+  log('tryMoveFocusFromClosingActiveTabOnFocusRedirected: tabs.onActivated is fired');
 
-  const autoFocusedTab = Tabs.getCurrentTab(tab.apiTab.windowId);
-  if (autoFocusedTab != nextTab &&
-      (autoFocusedTab != previousTab ||
-       (Tabs.getNextTab(autoFocusedTab) &&
-        Tabs.getNextTab(autoFocusedTab) != tab))) {
-    // possibly it is focused by browser.tabs.selectOwnerOnClose
-    log('=> the tab seems focused intentionally: ', {
-      autoFocused:       dumpTab(autoFocusedTab),
-      nextOfAutoFocused: dumpTab(Tabs.getNextTab(autoFocusedTab)),
+  const autoActiveTab = Tabs.getActiveTab(tab.apiTab.windowId);
+  if (autoActiveTab != nextTab &&
+      (autoActiveTab != previousTab ||
+       (Tabs.getNextTab(autoActiveTab) &&
+        Tabs.getNextTab(autoActiveTab) != tab))) {
+    // possibly it is active by browser.tabs.selectOwnerOnClose
+    log('=> the tab seems active intentionally: ', {
+      autoActive:       dumpTab(autoActiveTab),
+      nextOfAutoActive: dumpTab(Tabs.getNextTab(autoActiveTab)),
       prev:              dumpTab(previousTab),
       next:              dumpTab(nextTab)
     });
     return false;
   }
-  return tryMoveFocusFromClosingCurrentTabNow(tab, { params });
+  return tryMoveFocusFromClosingActiveTabNow(tab, { params });
 }
-function getTryMoveFocusFromClosingCurrentTabNowParams(tab, overrideParams) {
+function getTryMoveFocusFromClosingActiveTabNowParams(tab, overrideParams) {
   const parentTab = Tabs.getParentTab(tab);
   const params = {
     active:                    Tabs.isActive(tab),
@@ -836,8 +936,8 @@ function getTryMoveFocusFromClosingCurrentTabNowParams(tab, overrideParams) {
     firstChildTabOfParent:     Tabs.getFirstChildTab(parentTab),
     lastChildTabOfParent:      Tabs.getLastChildTab(parentTab),
     previousSiblingTab:        Tabs.getPreviousSiblingTab(tab),
-    preDetectedNextFocusedTab: Tabs.getNextFocusedTab(tab),
-    serialized:                TSTAPI.serializeTab(tab),
+    preDetectedNextActiveTab: Tabs.getNextActiveTab(tab),
+    serializedTab:             TSTAPI.serializeTab(tab),
     closeParentBehavior:       getCloseParentBehaviorForTab(tab, { parentTab })
   };
   if (overrideParams)
@@ -845,25 +945,25 @@ function getTryMoveFocusFromClosingCurrentTabNowParams(tab, overrideParams) {
   return params;
 }
 
-export async function tryMoveFocusFromClosingCurrentTabNow(tab, options = {}) {
-  if (!configs.moveFocusInTreeForClosedCurrentTab)
+export async function tryMoveFocusFromClosingActiveTabNow(tab, options = {}) {
+  if (configs.successorTabControlLevel != Constants.kSUCCESSOR_TAB_CONTROL_IN_TREE)
     return false;
-  const params = options.params || getTryMoveFocusFromClosingCurrentTabNowParams(tab);
+  const params = options.params || getTryMoveFocusFromClosingActiveTabNowParams(tab);
   if (options.ignoredTabs)
     params.ignoredTabs = options.ignoredTabs;
   const {
     active,
     nextTabUrl, nextIsDiscarded,
     parentTab, firstChildTab, firstChildTabOfParent, lastChildTabOfParent,
-    previousSiblingTab, preDetectedNextFocusedTab,
-    serialized, closeParentBehavior
+    previousSiblingTab, preDetectedNextActiveTab,
+    serializedTab, closeParentBehavior
   } = params;
   let {
     nextTab,
     ignoredTabs
   } = params;
 
-  log('tryMoveFocusFromClosingCurrentTabNow ', params);
+  log('tryMoveFocusFromClosingActiveTabNow ', params);
   if (!active) {
     log(' => not active tab');
     return false;
@@ -871,47 +971,48 @@ export async function tryMoveFocusFromClosingCurrentTabNow(tab, options = {}) {
 
   const results = await TSTAPI.sendMessage({
     type:   TSTAPI.kNOTIFY_TRY_MOVE_FOCUS_FROM_CLOSING_CURRENT_TAB,
-    tab:    serialized,
-    window: tab.apiTab.windowId
+    tab:    serializedTab,
+    window: tab.apiTab.windowId,
+    windowId: tab.apiTab.windowId
   });
   if (results.some(result => result.result)) // canceled
     return false;
 
-  let nextFocusedTab = null;
+  let nextActiveTab = null;
   if (firstChildTab &&
       (closeParentBehavior == Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN ||
        closeParentBehavior == Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD))
-    nextFocusedTab = firstChildTab;
-  log('focus to first child?: ', !!nextFocusedTab);
+    nextActiveTab = firstChildTab;
+  log('focus to first child?: ', !!nextActiveTab);
 
   ignoredTabs = ignoredTabs || [];
   if (parentTab) {
-    log(`tab=${dumpTab(tab)}, parent=${dumpTab(parentTab)}, nextFocused=${dumpTab(nextFocusedTab)}, lastChildTabOfParent=${dumpTab(lastChildTabOfParent)}, previousSiblingTab=${dumpTab(previousSiblingTab)}`);
-    if (!nextFocusedTab && tab == lastChildTabOfParent) {
+    log(`tab=${dumpTab(tab)}, parent=${dumpTab(parentTab)}, nextActive=${dumpTab(nextActiveTab)}, lastChildTabOfParent=${dumpTab(lastChildTabOfParent)}, previousSiblingTab=${dumpTab(previousSiblingTab)}`);
+    if (!nextActiveTab && tab == lastChildTabOfParent) {
       if (tab == firstChildTabOfParent) { // this is the really last child
-        nextFocusedTab = parentTab;
-        log('focus to parent?: ', !!nextFocusedTab);
+        nextActiveTab = parentTab;
+        log('focus to parent?: ', !!nextActiveTab);
       }
       else {
-        nextFocusedTab = previousSiblingTab;
-        log('focus to previous sibling?: ', !!nextFocusedTab);
+        nextActiveTab = previousSiblingTab;
+        log('focus to previous sibling?: ', !!nextActiveTab);
       }
     }
-    if (nextFocusedTab && ignoredTabs.includes(nextFocusedTab))
-      nextFocusedTab = Tabs.getNextFocusedTab(parentTab, { ignoredTabs });
+    if (nextActiveTab && ignoredTabs.includes(nextActiveTab))
+      nextActiveTab = Tabs.getNextActiveTab(parentTab, { ignoredTabs });
   }
-  else if (!nextFocusedTab) {
-    nextFocusedTab = preDetectedNextFocusedTab;
-    log('focus to Tabs.getNextFocusedTab()?: ', !!nextFocusedTab);
+  else if (!nextActiveTab) {
+    nextActiveTab = preDetectedNextActiveTab;
+    log('focus to Tabs.getNextActiveTab()?: ', !!nextActiveTab);
   }
-  if (nextFocusedTab && ignoredTabs.includes(nextFocusedTab)) {
-    nextFocusedTab = Tabs.getNextFocusedTab(nextFocusedTab, { ignoredTabs });
-    log('focus to Tabs.getNextFocusedTab() again?: ', !!nextFocusedTab);
+  if (nextActiveTab && ignoredTabs.includes(nextActiveTab)) {
+    nextActiveTab = Tabs.getNextActiveTab(nextActiveTab, { ignoredTabs });
+    log('focus to Tabs.getNextActiveTab() again?: ', !!nextActiveTab);
   }
 
-  if (!nextFocusedTab ||
-      Tabs.isHidden(nextFocusedTab) ||
-      Tabs.isActive(nextFocusedTab))
+  if (!nextActiveTab ||
+      Tabs.isHidden(nextActiveTab) ||
+      Tabs.isActive(nextActiveTab))
     return false;
 
   nextTab = Tabs.getTabById(nextTab);
@@ -921,8 +1022,8 @@ export async function tryMoveFocusFromClosingCurrentTabNow(tab, options = {}) {
     nextTab.dataset.discardURLAfterCompletelyLoaded = nextTabUrl || nextTab.apiTab.url;
   }
 
-  log('focus to: ', dumpTab(nextFocusedTab));
-  await TabsInternalOperation.selectTab(nextFocusedTab);
+  log('focus to: ', dumpTab(nextActiveTab));
+  await TabsInternalOperation.activateTab(nextActiveTab);
   return true;
 }
 
@@ -956,13 +1057,13 @@ export function getCloseParentBehaviorForTab(tab, options = {}) {
   return behavior;
 }
 
-export function getCloseParentBehaviorForTabWithSidebarOpenState(tab, info = {}) {
+export function getCloseParentBehaviorForTabWithSidebarOpenState(tab, removeInfo = {}) {
   return getCloseParentBehaviorForTab(tab, {
     keepChildren: (
-      info.keepChildren ||
+      removeInfo.keepChildren ||
       !shouldApplyTreeBehavior({
-        windowId:            info.windowId || tab.apiTab.windowId,
-        byInternalOperation: info.byInternalOperation
+        windowId:            removeInfo.windowId || tab.apiTab.windowId,
+        byInternalOperation: removeInfo.byInternalOperation
       })
     )
   });
@@ -1000,7 +1101,7 @@ export async function moveTabSubtreeBefore(tab, nextTab, options = {}) {
 
   log('moveTabSubtreeBefore: ', dumpTab(tab), dumpTab(nextTab));
   const container = tab.parentNode;
-  TabsContainer.incrementCounter(container, 'subTreeMovingCount');
+  container.subTreeMovingCount++;
   try {
     await TabsMove.moveTabInternallyBefore(tab, nextTab, options);
     if (!Tabs.ensureLivingTab(tab)) // it is removed while waiting
@@ -1013,7 +1114,7 @@ export async function moveTabSubtreeBefore(tab, nextTab, options = {}) {
   await wait(0);
   if (!container.parentNode) // it was removed while waiting
     return;
-  TabsContainer.decrementCounter(container, 'subTreeMovingCount');
+  container.subTreeMovingCount--;
 }
 
 export async function moveTabSubtreeAfter(tab, previousTab, options = {}) {
@@ -1026,7 +1127,7 @@ export async function moveTabSubtreeAfter(tab, previousTab, options = {}) {
 
   log('moveTabSubtreeAfter: ', dumpTab(tab), dumpTab(previousTab));
   const container = tab.parentNode;
-  TabsContainer.incrementCounter(container, 'subTreeMovingCount');
+  container.subTreeMovingCount++;
   try {
     await TabsMove.moveTabInternallyAfter(tab, previousTab, options);
     if (!Tabs.ensureLivingTab(tab)) // it is removed while waiting
@@ -1039,7 +1140,7 @@ export async function moveTabSubtreeAfter(tab, previousTab, options = {}) {
   await wait(0);
   if (!container.parentNode) // it was removed while waiting
     return;
-  TabsContainer.decrementCounter(container, 'subTreeMovingCount');
+  container.subTreeMovingCount--;
 }
 
 export async function followDescendantsToMovedRoot(tab, options = {}) {
@@ -1048,11 +1149,11 @@ export async function followDescendantsToMovedRoot(tab, options = {}) {
 
   log('followDescendantsToMovedRoot: ', dumpTab(tab));
   const container = tab.parentNode;
-  TabsContainer.incrementCounter(container, 'subTreeChildrenMovingCount');
-  TabsContainer.incrementCounter(container, 'subTreeMovingCount');
+  container.subTreeChildrenMovingCount++;
+  container.subTreeMovingCount++;
   await TabsMove.moveTabsAfter(Tabs.getDescendantTabs(tab), tab, options);
-  TabsContainer.decrementCounter(container, 'subTreeChildrenMovingCount');
-  TabsContainer.decrementCounter(container, 'subTreeMovingCount');
+  container.subTreeChildrenMovingCount--;
+  container.subTreeMovingCount--;
 }
 
 export async function moveTabs(tabs, options = {}) {
@@ -1103,9 +1204,11 @@ export async function moveTabs(tabs, options = {}) {
           Tabs.allTabsContainer.appendChild(container);
         }
         if (isAcrossWindows) {
-          TabsContainer.incrementCounter(container, 'toBeOpenedTabsWithPositions', tabs.length);
-          TabsContainer.incrementCounter(container, 'toBeOpenedOrphanTabs', tabs.length);
-          TabsContainer.incrementCounter(container, 'toBeAttachedTabs', tabs.length);
+          container.toBeOpenedTabsWithPositions += tabs.length;
+          container.toBeOpenedOrphanTabs += tabs.length;
+          for (const tab of tabs) {
+            container.toBeAttachedTabs.add(tab.apiTab.id);
+          }
         }
       };
       if (newWindow) {
@@ -1127,12 +1230,15 @@ export async function moveTabs(tabs, options = {}) {
         (async () => {
           const sourceContainer = tabs[0].parentNode;
           if (options.duplicate) {
-            TabsContainer.incrementCounter(sourceContainer, 'toBeOpenedTabsWithPositions', tabs.length);
-            TabsContainer.incrementCounter(sourceContainer, 'toBeOpenedOrphanTabs', tabs.length);
-            TabsContainer.incrementCounter(sourceContainer, 'duplicatingTabsCount', tabs.length);
+            sourceContainer.toBeOpenedTabsWithPositions += tabs.length;
+            sourceContainer.toBeOpenedOrphanTabs += tabs.length;
+            sourceContainer.duplicatingTabsCount += tabs.length;
+          }
+          if (isAcrossWindows) {
+            for (const tab of tabs) {
+              sourceContainer.toBeDetachedTabs.add(tab.apiTab.id);
+            }
           }
-          if (isAcrossWindows)
-            TabsContainer.incrementCounter(sourceContainer, 'toBeDetachedTabs', tabs.length);
 
           log('preparing tabs');
           if (options.duplicate) {
@@ -1214,11 +1320,13 @@ export async function moveTabs(tabs, options = {}) {
         toIndex--;
       log(' => ', toIndex);
       if (isAcrossWindows) {
-        for (const tab of tabs) {
-          if (!Tabs.isActive(tab))
-            continue;
-          await tryMoveFocusFromClosingCurrentTabNow(tab, { ignoredTabs: tabs });
-          break;
+        if (typeof browser.tabs.moveInSuccession != 'function') { // on Firefox 64 or older
+          for (const tab of tabs) {
+            if (!Tabs.isActive(tab))
+              continue;
+            await tryMoveFocusFromClosingActiveTabNow(tab, { ignoredTabs: tabs });
+            break;
+          }
         }
         apiTabs = await ApiTabs.safeMoveAcrossWindows(apiTabIds, {
           windowId: destinationWindowId,
@@ -1311,7 +1419,7 @@ export async function openNewWindowFromTabs(tabs, options = {}) {
 
   log('opening new window');
   const windowParams = {
-    //focused: true,  // not supported in Firefox...
+    //active: true,  // not supported in Firefox...
     url: 'about:blank',
     incognito: Tabs.isPrivateBrowsing(tabs[0])
   };
@@ -1352,204 +1460,91 @@ export async function openNewWindowFromTabs(tabs, options = {}) {
 }
 
 
-
-
-// drag and drop helper
-
-export async function performTabsDragDrop(params = {}) {
-  const windowId = params.windowId || Tabs.getWindow();
-  const destinationWindowId = params.destinationWindowId || windowId;
-
-  if (params.inRemote) {
-    browser.runtime.sendMessage(Object.assign({}, params, {
-      type:         Constants.kCOMMAND_PERFORM_TABS_DRAG_DROP,
-      windowId:     windowId,
-      attachTo:     params.attachTo && params.attachTo.id,
-      insertBefore: params.insertBefore && params.insertBefore.id,
-      insertAfter:  params.insertAfter && params.insertAfter.id,
-      inRemote:     false,
-      destinationWindowId
-    }));
-    return;
-  }
-
-  log('performTabsDragDrop ', {
-    tabs:                params.tabs.map(tab => tab.id),
-    windowId:            params.windowId,
-    destinationWindowId: params.destinationWindowId,
-    action:              params.action
-  });
-
-  let draggedTabs = params.tabs.map(Tabs.getTabById).filter(tab => !!tab);
-  if (!draggedTabs.length)
-    return;
-
-  // Basically tabs should not be dragged between regular window and private browsing window,
-  // so there are some codes to prevent shch operations. This is for failsafe.
-  if (Tabs.isPrivateBrowsing(draggedTabs[0]) != Tabs.isPrivateBrowsing(Tabs.getFirstTab(destinationWindowId)))
-    return;
-
-  let draggedRoots = Tabs.collectRootTabs(draggedTabs);
-
-  const draggedWholeTree = [].concat(draggedRoots);
-  for (const draggedRoot of draggedRoots) {
-    const descendants = Tabs.getDescendantTabs(draggedRoot);
-    for (const descendant of descendants) {
-      if (!draggedWholeTree.includes(descendant))
-        draggedWholeTree.push(descendant);
+export function calculateReferenceTabsFromInsertionPosition(tab, params = {}) {
+  if (params.insertBefore) {
+    /* strategy
+         +-----------------------------------------------------
+         |     <= detach from parent, and move
+         |[TARGET  ]
+         +-----------------------------------------------------
+         |  [      ]
+         |     <= attach to the parent of the target, and move
+         |[TARGET  ]
+         +-----------------------------------------------------
+         |[        ]
+         |     <= attach to the parent of the target, and move
+         |[TARGET  ]
+         +-----------------------------------------------------
+         |[        ]
+         |     <= attach to the parent of the target (previous tab), and move
+         |  [TARGET]
+         +-----------------------------------------------------
+    */
+    const prevTab = Tabs.getPreviousVisibleTab(params.insertBefore);
+    if (!prevTab) {
+      // allow to move pinned tab to beside of another pinned tab
+      if (!tab || Tabs.isPinned(tab) == Tabs.isPinned(params.insertBefore)) {
+        return {
+          insertBefore: params.insertBefore
+        };
+      }
+      else {
+        return {};
+      }
+    }
+    else {
+      const prevLevel   = Number(prevTab.getAttribute(Constants.kLEVEL) || 0);
+      const targetLevel = Number(params.insertBefore.getAttribute(Constants.kLEVEL) || 0);
+      let parent = null;
+      if (!tab || !Tabs.isPinned(tab))
+        parent = (prevLevel < targetLevel) ? prevTab : Tabs.getParentTab(params.insertBefore);
+      return {
+        parent,
+        insertAfter:  prevTab,
+        insertBefore: params.insertBefore
+      }
     }
   }
-  log('=> draggedTabs: ', draggedTabs.map(dumpTab).join(' / '));
-
-  if (draggedWholeTree.length != draggedTabs.length) {
-    log('=> partially dragged');
-    if (!params.duplicate)
-      await detachTabsFromTree(draggedTabs, {
-        broadcast: true
-      });
-  }
-
-  while (params.insertBefore &&
-         draggedWholeTree.includes(params.insertBefore)) {
-    params.insertBefore = Tabs.getNextTab(params.insertBefore);
-  }
-  while (params.insertAfter &&
-         draggedWholeTree.includes(params.insertAfter)) {
-    params.insertAfter = Tabs.getPreviousTab(params.insertAfter);
-  }
-
-  if (params.duplicate ||
-      windowId != destinationWindowId) {
-    draggedTabs = await moveTabs(draggedTabs, {
-      destinationWindowId,
-      duplicate:    params.duplicate,
-      insertBefore: params.insertBefore,
-      insertAfter:  params.insertAfter
-    });
-    draggedRoots = Tabs.collectRootTabs(draggedTabs);
-  }
-
-  log('try attach/detach');
-  if (!params.attachTo) {
-    log('=> detach');
-    detachTabsOnDrop(draggedRoots, {
-      broadcast: true
-    });
-  }
-  else if (params.action & Constants.kACTION_ATTACH) {
-    log('=> attach');
-    await attachTabsOnDrop(draggedRoots, params.attachTo, {
-      insertBefore: params.insertBefore,
-      insertAfter:  params.insertAfter,
-      draggedTabs:  draggedTabs,
-      broadcast:    true
-    });
-  }
-  else {
-    log('=> just moved');
-  }
-
-  log('=> moving dragged tabs ', draggedTabs.map(dumpTab));
-  if (params.insertBefore)
-    await TabsMove.moveTabsBefore(draggedTabs, params.insertBefore);
-  else if (params.insertAfter)
-    await TabsMove.moveTabsAfter(draggedTabs, params.insertAfter);
-  else
-    log('=> already placed at expected position');
-
-  if (windowId != destinationWindowId) {
-    // Firefox always focuses to the dropped tab if it is dragged from another window.
-    // TST respects Firefox's the behavior.
-    browser.tabs.update(draggedTabs[0].apiTab.id, { active: true })
-      .catch(ApiTabs.handleMissingTabError);
-  }
-
-  /*
-  const treeStructure = getTreeStructureFromTabs(draggedTabs);
-
-  const newTabs;
-  const replacedGroupTabs = Tabs.doAndGetNewTabs(() => {
-    newTabs = moveTabsInternal(draggedTabs, {
-      duplicate    : params.duplicate,
-      insertBefore : params.insertBefore,
-      insertAfter  : params.insertAfter,
-      inRemote     : true
-    });
-  });
-  log('=> opened group tabs: ', replacedGroupTabs);
-  params.draggedTab.ownerDocument.defaultView.setTimeout(() => {
-    if (!Tabs.ensureLivingTab(tab)) // it was removed while waiting
-      return;
-    log('closing needless group tabs');
-    replacedGroupTabs.reverse().forEach(function(tab) {
-      log(' check: ', tab.label+'('+tab._tPos+') '+getLoadingURI(tab));
-      if (Tabs.isGroupTab(tab) &&
-        !Tabs.hasChildTabs(tab))
-        removeTab(tab);
-    }, this);
-  }, 0);
-  */
-
-  /*
-  if (newTabs.length && params.action & Constants.kACTION_ATTACH) {
-    Promise.all(newTabs.map((tab) => tab.__treestyletab__promisedDuplicatedTab))
-      .then((function() {
-        log('   => attach (last)');
-        await attachTabsOnDrop(
-          newTabs.filter(function(tab, index) {
-            return treeStructure[index] == -1;
-          }),
-          params.attachTo,
-          { insertBefore: params.insertBefore,
-            insertAfter:  params.insertAfter }
-        );
-      }).bind(this));
-  }
-  */
-
-  log('=> finished');
-}
-
-async function attachTabsOnDrop(tabs, parent, options = {}) {
-  log('attachTabsOnDrop: start ', tabs.map(dumpTab));
-  if (parent && !options.insertBefore && !options.insertAfter) {
-    const refTabs = getReferenceTabsForNewChild(tabs[0], parent, {
-      ignoreTabs: tabs
-    });
-    options.insertBefore = refTabs.insertBefore;
-    options.insertAfter  = refTabs.insertAfter;
-  }
-
-  if (options.insertBefore)
-    await TabsMove.moveTabsBefore(options.draggedTabs || tabs, options.insertBefore);
-  else if (options.insertAfter)
-    await TabsMove.moveTabsAfter(options.draggedTabs || tabs, options.insertAfter);
-
-  const memberOptions = Object.assign({}, options, {
-    insertBefore: null,
-    insertAfter:  null,
-    dontMove:     true,
-    forceExpand:  options.draggedTabs.some(Tabs.isActive)
-  });
-  for (const tab of tabs) {
-    if (parent)
-      attachTabTo(tab, parent, memberOptions);
-    else
-      detachTab(tab, memberOptions);
-    collapseExpandTabAndSubtree(tab, Object.assign({}, memberOptions, {
-      collapsed: false
-    }));
-  }
-}
-
-function detachTabsOnDrop(tabs, options = {}) {
-  log('detachTabsOnDrop: start ', tabs.map(dumpTab));
-  for (const tab of tabs) {
-    detachTab(tab, options);
-    collapseExpandTabAndSubtree(tab, Object.assign({}, options, {
-      collapsed: false
-    }));
+  if (params.insertAfter) {
+    /* strategy
+         +-----------------------------------------------------
+         |[TARGET  ]
+         |     <= if the target has a parent, attach to it and and move
+         +-----------------------------------------------------
+         |  [TARGET]
+         |     <= attach to the parent of the target, and move
+         |[        ]
+         +-----------------------------------------------------
+         |[TARGET  ]
+         |     <= attach to the parent of the target, and move
+         |[        ]
+         +-----------------------------------------------------
+         |[TARGET  ]
+         |     <= attach to the target, and move
+         |  [      ]
+         +-----------------------------------------------------
+    */
+    const nextTab = Tabs.getNextVisibleTab(params.insertAfter);
+    if (!nextTab) {
+      return {
+        parent:      Tabs.getParentTab(params.insertAfter),
+        insertAfter: params.insertAfter
+      };
+    }
+    else {
+      const targetLevel = Number(params.insertAfter.getAttribute(Constants.kLEVEL) || 0);
+      const nextLevel   = Number(nextTab.getAttribute(Constants.kLEVEL) || 0);
+      let parent = null;
+      if (!tab || !Tabs.isPinned(tab))
+        parent = (targetLevel < nextLevel) ? params.insertAfter : Tabs.getParentTab(params.insertAfter) ;
+      return {
+        parent,
+        insertBefore: nextTab,
+        insertAfter:  params.insertAfter
+      };
+    }
   }
+  throw new Error('calculateReferenceTabsFromInsertionPosition requires one of insertBefore or insertAfter parameter!');
 }
 
 
