@@ -55,6 +55,7 @@ function log(...args) {
 
 export const kREGISTER_SELF         = 'register-self';
 export const kUNREGISTER_SELF       = 'unregister-self';
+export const kWAIT_FOR_SHUTDOWN     = 'wait-for-shutdown';
 export const kPING                  = 'ping';
 export const kNOTIFY_READY          = 'ready';
 export const kNOTIFY_SHUTDOWN       = 'shutdown'; // defined but not notified for now.
@@ -75,6 +76,8 @@ export const kNOTIFY_TAB_DRAGSTART  = 'tab-dragstart';
 export const kNOTIFY_TAB_DRAGENTER  = 'tab-dragenter';
 export const kNOTIFY_TAB_DRAGEXIT   = 'tab-dragexit';
 export const kNOTIFY_TAB_DRAGEND    = 'tab-dragend';
+export const kNOTIFY_NATIVE_TAB_DRAGSTART = 'native-tab-dragstart';
+export const kSTART_CUSTOM_DRAG     = 'start-custom-drag';
 export const kNOTIFY_TRY_MOVE_FOCUS_FROM_CLOSING_CURRENT_TAB = 'try-move-focus-from-closing-current-tab';
 export const kGET_TREE              = 'get-tree';
 export const kATTACH                = 'attach';
@@ -84,10 +87,14 @@ export const kDEMOTE                = 'demote';
 export const kOUTDENT               = 'outdent';
 export const kPROMOTE               = 'promote';
 export const kMOVE_UP               = 'move-up';
+export const kMOVE_TO_START         = 'move-to-start';
 export const kMOVE_DOWN             = 'move-down';
+export const kMOVE_TO_END           = 'move-to-end';
 export const kFOCUS                 = 'focus';
 export const kDUPLICATE             = 'duplicate';
 export const kGROUP_TABS            = 'group-tabs';
+export const kOPEN_IN_NEW_WINDOW    = 'open-in-new-window';
+export const kREOPEN_IN_CONTAINER   = 'reopen-in-container';
 export const kGET_TREE_STRUCTURE    = 'get-tree-structure';
 export const kSET_TREE_STRUCTURE    = 'set-tree-structure';
 export const kCOLLAPSE_TREE         = 'collapse-tree';
@@ -100,6 +107,7 @@ export const kSCROLL_UNLOCK         = 'scroll-unlock';
 export const kNOTIFY_SCROLLED       = 'scrolled';
 export const kBLOCK_GROUPING        = 'block-grouping';
 export const kUNBLOCK_GROUPING      = 'unblock-grouping';
+export const kGRANT_TO_REMOVE_TABS  = 'grant-to-remove-tabs';
 
 export const kCONTEXT_MENU_UPDATED    = 'fake-contextMenu-updated';
 export const kCONTEXT_MENU_GET_ITEMS  = 'fake-contextMenu-get-items';
@@ -132,10 +140,35 @@ export function getAddon(id) {
 }
 
 function registerAddon(id, addon) {
+  log('addon is registered: ', id, addon);
+
+  // inherit properties from last effective value
+  const oldAddon = mAddons.get(id);
+  if (oldAddon) {
+    if (!('listeningTypes' in addon) && 'listeningTypes' in oldAddon)
+      addon.listeningTypes = oldAddon.listeningTypes;
+    if (!('style' in addon) && 'style' in oldAddon)
+      addon.style = oldAddon.style;
+  }
+
+  if (!addon.listeningTypes) {
+    // for backward compatibility, send all message types available on TST 2.4.16 by default.
+    addon.listeningTypes = [
+      kNOTIFY_READY,
+      kNOTIFY_SHUTDOWN,
+      kNOTIFY_TAB_CLICKED,
+      kNOTIFY_TAB_MOUSEDOWN,
+      kNOTIFY_TAB_MOUSEUP,
+      kNOTIFY_TABBAR_CLICKED,
+      kNOTIFY_TABBAR_MOUSEDOWN,
+      kNOTIFY_TABBAR_MOUSEUP
+    ];
+  }
   mAddons.set(id, addon);
 }
 
 function unregisterAddon(id) {
+  log('addon is unregistered: ', id, mAddons.get(id));
   mAddons.delete(id);
   delete mScrollLockedBy[id];
   delete mGroupingBlockedBy[id];
@@ -157,24 +190,32 @@ export async function initAsBackend() {
   const respondedAddons = [];
   const notifiedAddons = {};
   const notifyAddons = configs.knownExternalAddons.concat(configs.cachedExternalAddons);
+  log('initAsBackend: notifyAddons = ', respondedAddons);
   await Promise.all(notifyAddons.map(async id => {
     if (id in notifiedAddons)
       return;
     notifiedAddons[id] = true;
     try {
-      const success = await browser.runtime.sendMessage(id, {
-        type: kNOTIFY_READY
+      id = await new Promise((resolve, reject) => {
+        browser.runtime.sendMessage(id, {
+          type: kNOTIFY_READY
+        }).then(() => resolve(id)).catch(reject);
+        setTimeout(() => {
+          reject(new Error(`TSTAPI.initAsBackend: addon ${id} does not respond.`));
+        }, 3000);
       });
-      if (success)
+      if (id)
         respondedAddons.push(id);
     }
-    catch(_e) {
+    catch(e) {
+      console.log(e);
     }
   }));
+  log('initAsBackend: respondedAddons = ', respondedAddons);
   configs.cachedExternalAddons = respondedAddons;
 }
 
-browser.runtime.onMessage.addListener((message, _aSender) => {
+browser.runtime.onMessage.addListener((message, _sender) => {
   if (!message ||
       typeof message.type != 'string')
     return;
@@ -213,6 +254,12 @@ browser.runtime.onMessage.addListener((message, _aSender) => {
   }
 });
 
+const mPromisedOnBeforeUnload = new Promise((resolve, _reject) => {
+  // If this promise doesn't do anything then there seems to be a timeout so it only works if TST is disabled within about 10 seconds after this promise is used as a response to a message. After that it will not throw an error for the waiting extension.
+  // If we use the following then the returned promise will be rejected when TST is disabled even for longer times:
+  window.addEventListener('beforeunload', () => resolve());
+});
+
 browser.runtime.onMessageExternal.addListener((message, sender) => {
   if (!message ||
       typeof message.type != 'string')
@@ -227,19 +274,6 @@ browser.runtime.onMessageExternal.addListener((message, sender) => {
 
         case kREGISTER_SELF:
           return (async () => {
-            if (!message.listeningTypes) {
-              // for backward compatibility, send all message types available on TST 2.4.16 by default.
-              message.listeningTypes = [
-                kNOTIFY_READY,
-                kNOTIFY_SHUTDOWN,
-                kNOTIFY_TAB_CLICKED,
-                kNOTIFY_TAB_MOUSEDOWN,
-                kNOTIFY_TAB_MOUSEUP,
-                kNOTIFY_TABBAR_CLICKED,
-                kNOTIFY_TABBAR_MOUSEDOWN,
-                kNOTIFY_TABBAR_MOUSEUP
-              ];
-            }
             message.internalId = sender.url.replace(/^moz-extension:\/\/([^\/]+)\/.*$/, '$1');
             message.id = sender.id;
             registerAddon(sender.id, message);
@@ -266,6 +300,9 @@ browser.runtime.onMessageExternal.addListener((message, sender) => {
             configs.cachedExternalAddons = configs.cachedExternalAddons.filter(id => id != sender.id);
             return true;
           })();
+
+        case kWAIT_FOR_SHUTDOWN:
+          return mPromisedOnBeforeUnload;
       }
       break;
 
@@ -367,6 +404,7 @@ export async function notifyScrolled(params = {}) {
     tab:  tab && serializeTab(tab),
     tabs: Tabs.getTabs(window).map(serializeTab),
     window,
+    windowId: window,
 
     deltaY:       params.event.deltaY,
     deltaMode:    params.event.deltaMode,
@@ -396,17 +434,29 @@ export function isGroupingBlocked() {
 
 
 export function serializeTab(tab) {
-  const effectiveFavIcon = TabFavIconHelper.effectiveFavIcons.get(tab.apiTab.id);
   const children         = Tabs.getChildTabs(tab).map(serializeTab);
   const ancestorTabIds   = Tabs.getAncestorTabs(tab).map(tab => tab.apiTab.id);
   return Object.assign({}, tab.apiTab, {
-    states:   Array.slice(tab.classList).filter(state => !Constants.kTAB_INTERNAL_STATES.includes(state)),
+    states:   Array.from(tab.classList).filter(state => !Constants.kTAB_INTERNAL_STATES.includes(state)),
     indent:   parseInt(tab.getAttribute(Constants.kLEVEL) || 0),
-    effectiveFavIconUrl: effectiveFavIcon && effectiveFavIcon.favIconUrl,
     children, ancestorTabIds
   });
 }
 
+export async function serializeTabWithEffectiveFavIconUrl(tab) {
+  const serializedRoot = serializeTab(tab);
+  const promises = [];
+  const preparePromiseForEffectiveFavIcon = serializedOneTab => {
+    promises.push(TabFavIconHelper.getLastEffectiveFavIconURL(serializedOneTab).then(url => {
+      serializedOneTab.effectiveFavIconUrl = url;
+    }));
+    serializedOneTab.children.map(preparePromiseForEffectiveFavIcon);
+  };
+  preparePromiseForEffectiveFavIcon(serializedRoot);
+  await Promise.all(promises);
+  return serializedRoot;
+}
+
 export function getListenersForMessageType(type) {
   const uniqueTargets = {};
   for (const [id, addon] of getAddons()) {
@@ -418,9 +468,15 @@ export function getListenersForMessageType(type) {
 
 export async function sendMessage(message, options = {}) {
   const uniqueTargets = new Set();
-  for (const addon of getListenersForMessageType(message.type)) {
+  const listenerAddons = getListenersForMessageType(message.type);
+  for (const addon of listenerAddons) {
     uniqueTargets.add(addon.id);
   }
+  log(`sendMessage: sending message for ${message.type}: `, {
+    message,
+    listenerAddons,
+    targets: options.targets
+  });
   if (options.targets) {
     if (!Array.isArray(options.targets))
       options.targets = [options.targets];
@@ -430,7 +486,10 @@ export async function sendMessage(message, options = {}) {
   }
 
   const promisedResults = spawnMessages(uniqueTargets, message);
-  return Promise.all(promisedResults);
+  return Promise.all(promisedResults).then(results => {
+    log(`sendMessage: got responses for ${message.type}: `, results);
+    return results;
+  });
 }
 
 function* spawnMessages(targetSet, message) {
@@ -457,7 +516,7 @@ function* spawnMessages(targetSet, message) {
 
 
 export async function getTargetTabs(message, sender) {
-  await Tabs.waitUntilAllTabsAreCreated();
+  await Tabs.waitUntilAllTabsAreCreated(message.window || message.windowId);
   if (Array.isArray(message.tabs))
     return getTabsFromWrongIds(message.tabs, sender);
   if (message.window || message.windowId) {
@@ -480,6 +539,7 @@ export async function getTargetTabs(message, sender) {
 }
 
 async function getTabsFromWrongIds(aIds, sender) {
+  log('getTabsFromWrongIds ', aIds, sender);
   let tabsInActiveWindow = [];
   if (aIds.some(id => typeof id != 'number')) {
     const window = await browser.windows.getLastFocused({
@@ -519,12 +579,25 @@ async function getTabsFromWrongIds(aIds, sender) {
       case 'sendertab':
         if (sender.tab)
           return sender.tab;
+      case 'highlighted':
+      case 'multiselected': {
+        const tabs = tabsInActiveWindow.filter(tab => tab.highlighted);
+        return tabs.map(tab => TabIdFixer.fixTab(tab));
+      }
       default:
         const tabFromUniqueId = Tabs.getTabByUniqueId(id);
         return tabFromUniqueId || id;
     }
   }));
-  return tabOrAPITabOrIds.map(Tabs.getTabById).filter(tab => !!tab);
+  log('=> ', tabOrAPITabOrIds);
+
+  let flattenTabOrAPITabOrIds;
+  if (typeof Array.prototype.flat !== 'function') // Firefox 61 or older
+    flattenTabOrAPITabOrIds = tabOrAPITabOrIds.reduce((acc, val) => acc.concat(val), []);
+  else
+    flattenTabOrAPITabOrIds = tabOrAPITabOrIds.flat();
+
+  return flattenTabOrAPITabOrIds.map(Tabs.getTabById).filter(tab => !!tab);
 }
 
 export function formatResult(results, originalMessage) {
